#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import division

import numpy as np
import argparse
from matplotlib import use as mplt_use
mplt_use('Agg')
import matplotlib.pyplot as plt

import hicexplorer.HiCMatrix as HiCMatrix
from hicexplorer.utilities import remove_outliers


def parse_arguments(args=None):
    parser = argparse.ArgumentParser(
        description='This program makes a distance vs. hi-c counts plot per chromosome.')

    # define the arguments
    parser.add_argument('--matrix', '-m',
                        help='HiC normalized (corrected) matrix',
                        required=True)

    parser.add_argument('--skipDiagonal', '-s',
                        help='If set, diagonal counts are not included',
                        action='store_true')

    parser.add_argument('--plotFile', '-o',
                        help='File name to save the file. The given file '
                        'ending will be used '
                        'to determine the image format. '
                        'The available options are: .png, .emf, '
                        '.eps, .pdf and .svg.',
                        type=argparse.FileType('w'),
                        metavar='file name',
                        required=True)

    parser.add_argument('--chromosomeExclude',
                        help='Exclude the given list of chrosomes from ',
                        nargs='+')

    parser.add_argument('--plotTitle', '-t',
                        help='Plot title')

    return parser


def chr_diagonals(matrix_file_name, chromosome_exclude):
    """
    extract the counts by distance per chromosome
    """
    hic_ma = HiCMatrix.hiCMatrix(matrix_file_name)
    print "removing unwanted chrs"
    hic_ma.filterUnwantedChr()
    if chromosome_exclude is None:
        chromosome_exclude = []

    chrtokeep = [x for x in hic_ma.interval_trees.keys() if x not in chromosome_exclude]
    print "Number of contacts {}".format(hic_ma.matrix.sum())
    hic_ma.keepOnlyTheseChr(chrtokeep)
    diagonal_dict = hic_ma.getCountsByDistance(per_chr=True)

    common_dist = None
    max_dist = 0
    chrom_list = hic_ma.getChrNames()
    for chrom in chrom_list:
        try:
            distances = diagonal_dict[chrom].keys()
            distances[0]
        except (KeyError, IndexError):
            continue
        # get list of common distances
        if max(distances) > max_dist:
            max_dist = max(distances)
        if common_dist is None:
            common_dist = set(distances)
        else:
            common_dist = common_dist.intersection(distances)

    return diagonal_dict, chrom_list, list(common_dist), max_dist


def main(args=None):
    """
    for each distance, compare the
    distribution of two samples,
    report number of cases were they differ
    """
    args = parse_arguments().parse_args(args)

    diagonals_dict, chrom_list, common_dist, max_dist = chr_diagonals(args.matrix, args.chromosomeExclude)
    plt.figure(figsize=(5, 4))

    max_mean = 0
    min_mean = 1e6
    dist_list = np.sort(common_dist)
    for chrom in chrom_list:
        values = diagonals_dict[chrom]
        if len(values):
            mean_values = [remove_outliers(values[dist]).mean() for dist in dist_list]
            prc_values_90 = [np.percentile(remove_outliers(values[dist]), 90) for dist in dist_list]

            if max(prc_values_90) > max_mean:
                max_mean = max(prc_values_90)
            if min(mean_values) < min_mean:
                min_mean = min(mean_values)
            label = chrom
            plt.plot(dist_list, mean_values, label=label)

    plt.yscale('log')
    plt.xscale('log')
    plt.legend(prop={'size': 'small'})
    plt.suptitle(args.plotTitle)
    plt.xlabel('genomic distance')
    plt.ylabel('corrected Hi-C counts')
    plt.xlim(min(dist_list), 1e6)
    plt.ylim(1, max_mean + max_mean * 0.1)
    plt.savefig(args.plotFile.name, bbox_inches='tight')



import sys
import collections


class ReadBed(object):
    """
    Reads a bed file. Based on the number of fields
    it tries to guess the type of bed file used. Current options
    are bed3, bed6 and bed12

    Example:
    bed = readBed(open("file.bed", 'r'))
    for interval in bed:
        print bed['start']

    """

    def __init__(self, file_handle):
        """
        :param file_handle: file handle
        :return:
        """

        self.file_type = None
        self.file_handle = file_handle
        self.line_number = 0
        # guess file type
        fields = self.get_no_comment_line().split('\t')
        self.guess_file_type(fields)
        self.file_handle.seek(0)
        self.prev_chrom = None
        self.prev_start = -1
        self.prev_line = None

        # list of bed fields
        self.fields = ['chromosome', 'start', 'end',
                       'name', 'score', 'strand',
                       'thick_start', 'thick_end',
                       'rgb', 'block_count',
                       'block_sizes', 'block_starts']

        if self.file_type == 'bed12':
            self.BedInterval = collections.namedtuple('BedInterval', self.fields)
        elif self.file_type == 'bed9':
            self.BedInterval = collections.namedtuple('BedInterval', self.fields[:9])
        else:
            self.BedInterval = collections.namedtuple('BedInterval', self.fields[:6])

    def __iter__(self):
        return self

    def get_no_comment_line(self):
        """
        Skips comment lines starting with '#'
        "track" or "browser" in the bed files
        :return:
        """
        line = self.file_handle.next()
        if line.startswith("#") or line.startswith("track") or \
           line.startswith("browser") or line.strip() == '':
            line = self.get_no_comment_line()

        self.line_number += 1
        return line

    def guess_file_type(self, line_values):
        """try to guess type of bed file by counting the fields
        """
        if len(line_values) == 3:
            self.file_type = 'bed3'
        elif len(line_values) == 4:
            self.file_type = 'bedgraph'
        elif len(line_values) == 6:
            self.file_type = 'bed6'
        elif len(line_values) == 12:
            self.file_type = 'bed12'
        elif len(line_values) == 9:
            # this is a case where a specific color is encoded in the 10 field of the bed file
            self.file_type = 'bed9'
        elif len(line_values) > 6:
            # assume bed6
            self.file_type = 'bed6'
            sys.stderr.write("Number of fields in BED file is not standard. Assuming bed6\n")
        else:
            # assume bed3
            self.file_type = 'bed3'
            sys.stderr.write("Number of fields in BED file is not standard. Assuming bed3\n")
        return self.file_type

    def next(self):
        """
        :return: bedInterval object
        """
        line = self.get_no_comment_line()

        bed = self.get_bed_interval(line)
        if self.prev_chrom == bed.chromosome:
            assert self.prev_start <= bed.start, \
                "Bed file not sorted. Please use a sorted bed file.\n" \
                "File: {}\n" \
                "Previous line: {}\n Current line{} ".format(self.file_handle.name, self.prev_line, line)

        self.prev_chrom = bed.chromosome
        self.prev_start = bed.start
        self.prev_line = line

        return bed

    def get_bed_interval(self, bed_line):
        r"""
        Processes each bed line from a bed file, casts the values and returns
        a namedtuple object

        >>> bed_line="chr1\t0\t1000\tgene_1\t0.5\t-\t0\t1000\t0\t3\t10,20,100\t20,200,700"
        >>> with open('/tmp/test.bed', 'w') as fh:
        ...     fh.write(bed_line)
        >>> bed_f = ReadBed(open('/tmp/test.bed','r'))
        >>> bed = bed_f.get_bed_interval(bed_line)
        >>> bed.chromosome
        'chr1'
        >>> bed.block_starts
        [20, 200, 700]

        >>> bed_line="chr2\t0\t1000\tgene_1\t0.5\t-\n"
        >>> with open('/tmp/test.bed', 'w') as fh:
        ...     fh.write(bed_line)
        >>> bed_f = ReadBed(open('/tmp/test.bed','r'))
        >>> bed_f.get_bed_interval(bed_line)
        BedInterval(chromosome='chr2', start=0, end=1000, name='gene_1', score=0.5, strand='-')
        """

        line_data = bed_line.strip().split("\t")
        if self.file_handle == 'bed12':
            assert len(line_data) == 12, "File type detected is bed12 but line {}: {} does " \
                                         "not have 12 fields.".format(self.line_number, bed_line)

        elif self.file_handle == 'bed3':
            assert len(line_data) == 3, "File type detected is bed3 but line {}: {} does " \
                "not have 3 fields.".format(self.line_number, bed_line)

        elif self.file_handle == 'bed6':
            assert len(line_data) == 6, "File type detected is bed6 but line {}: {} does " \
                "not have 3 fields.".format(self.line_number, bed_line)
        line_values = []
        for idx, r in enumerate(line_data):
            # first field is always chromosome/contig name
            # and should be cast as a string
            # same for field 3 (name)
            if idx in [0, 3]:
                line_values.append(r)
            # check field strand
            elif idx == 5:
                if r not in ['+', '-', '.']:
                    if r == '1':
                        r = '+'
                    elif r == '-1':
                        r = '-'
                    else:
                        sys.stderr.write("*Warning, invalid strand value found {} for line #{}:\n{}\n "
                                         "Setting strand to '.'\n".format(r, bed_line, self.line_number))
                        r = '.'
                line_values.append(r)

            elif idx in [1, 2, 6, 7, 9]:
                # start and end fields must be integers, same for thichStart(6),
                # and thickEnd(7) and blockCount(9) fields
                try:
                    line_values.append(int(r))
                except ValueError:
                    sys.stderr.write("Value: {} in field {} at line {} is not an integer\n".format(r, idx + 1,
                                                                                                   self.line_number))
                    return dict()
            # check item rgb
            elif idx == 8:
                rgb = r.split(",")
                if len(rgb) == 3:
                    try:
                        r = map(int, rgb)
                    except ValueError as detail:
                        sys.stderr.write("Error reading line: #{}. The rgb field {} is not "
                                         "valid.\nError message: {}\n".format(self.line_number, r, detail))
                line_values.append(r)

            elif idx in [10, 11]:
                # this are the block sizes and block start positions
                r_parts = r.split(',')
                try:
                    r = [int(x) for x in r_parts if x != '']
                except ValueError as detail:
                    sys.stderr.write("Error reading line #{}. The block field {} is not "
                                     "valid.\nError message: {}\n".format(self.line_number, r, detail))
                line_values.append(r)

            else:
                try:
                    tmp = float(r)
                except ValueError:
                    tmp = r
                except TypeError:
                    tmp = r
                line_values.append(tmp)

        assert line_values[2] > line_values[1], \
            "Start position larger or equal than end for line #{}:\n{}\n".format(self.line_number,
                                                                                 bed_line)

        if self.file_type == 'bed3':
            line_values = line_values[0:3]
            # in case of a bed3, the id, score and strand
            # values are added as ".", 0, "." respectively
            line_values.extend([".", 0, "."])
        elif self.file_type == 'bed6':
            line_values = line_values[0:6]

        return self.BedInterval._make(line_values)



from __future__ import division
import sys
import time
import numpy as np
import scipy.stats
import scipy.sparse
import argparse
from matplotlib import use as mplt_use
mplt_use('Agg')

from hicexplorer.HiCMatrix import hiCMatrix


def writableFile(string):
    try:
        open(string, 'w').close()
    except:
        msg = "{} file can be opened for writing".format(string)
        raise argparse.ArgumentTypeError(msg)
    return string


def mylog(data):
    return np.log(data + 1)


def getZscores(hicma):
    return transformMatrix(hicma, 'z-score')


def getTscores(hicma):
    return transformMatrix(hicma, 't-score')


def getResiduals(hicma):
    return transformMatrix(hicma, 'residuals')


def getObsExp(hicma):
    return transformMatrix(hicma, 'obs/exp')


def getPearson(matrix):
    matrix = convertNansToZeros(matrix).todense()
    from scipy.stats import pearsonr
    numRows, numCols = matrix.shape
    # create matrix to hold computed pval
    pMa = np.zeros(shape=(numCols, numRows))
    pMa[:, :] = np.nan
    for row in range(numRows):
        if row % 10 == 0:
            sys.stderr.write("{} rows processed ({:.2f})\n".format(row, float(row) / numRows))
        for col in range(numCols):
            if not np.isnan(pMa[col, row]):
                pMa[row, col] = pMa[col, row]
                continue
            try:
                # pearsonr returns two values, the first is the
                # correlation, the second is a pvalue.
                pMa[row, col] = pearsonr(np.asarray(matrix[row, :])[0], np.asarray(matrix[:, col].T)[0])[0]
            except:
                continue

    return pMa


def transformMatrix(hicma, method, per_chr=False, original_matrix=None, depth_in_bins=None):
    methods_avail = {'residuals': _residuals, 'obs/exp': _obsExp,
                     'z-score': _zscore, 't-score': _tscore,
                     'p-value': _pvalue, 'nbinom-p-value': _nbinomPvalue,
                     'nbinom-expected': _nbinomExpected,
                     'log-norm': _lognormPvalue,
                     'chi-squared': _chi2Pvalue}

    counts_by_distance, cut_intervals = hicma.getCountsByDistance(per_chr=per_chr, depth_in_bins=depth_in_bins)
    if method in ['nbinom-p-value', 'nbinom-expected', 'nbinom-est-dist']:
        size, prob = fitNegBinom_Rserve(counts_by_distance, per_chr=per_chr,
                                        plot_distribution=True)
    elif method == 'log-norm':
        mu_, sigma = fitDistribution(counts_by_distance, 'lognorm')
    else:
        if per_chr:
            mu_ = {}
            sigma = {}
            n_value = {}
            for chrom in counts_by_distance.keys():
                mu_[chrom] = dict([(x, np.mean(counts_by_distance[chrom][x]))
                                   for x in counts_by_distance[chrom]])
                sigma[chrom] = dict([(x, np.std(counts_by_distance[chrom][x]))
                                     for x in counts_by_distance[chrom]])
                n_value[chrom] = dict([(x, len(counts_by_distance[chrom][x]))
                                       for x in counts_by_distance[chrom]])
        else:
            mu_ = dict([(x, np.mean(counts_by_distance[x]))
                        for x in counts_by_distance])
            sigma = dict([(x, np.std(counts_by_distance[x]))
                          for x in counts_by_distance])
            n_value = dict([(x, len(counts_by_distance[x]))
                            for x in counts_by_distance])

    # use only the upper half of the matrix
    triu_ma = scipy.sparse.triu(hicma.matrix, format='coo')
    if original_matrix:
        orig_ma = original_matrix.matrix
        if per_chr:
            noise_level = {}
            for chrom in counts_by_distance.keys():
                chr_range = original_matrix.getChrBinRange(chrom)
                chr_submatrix = orig_ma[chr_range[0]:chr_range[1],
                                        chr_range[0]:chr_range[1]]
                noise_level[chrom] = np.median(chr_submatrix.data)
        else:
            noise_level = np.median(orig_ma.data)

        sys.stderr.write('noise error set to {}\n'.format(noise_level))
    else:
        noise_level = None
    sys.stderr.write("finish computing fitting parameters\n")

    ########################
    # after the distributions are fitted
    # now the matrix values are evaluated
    if method == 'nbinom-est-dist':
        triu_ma = nbinom_est_dist(size, prob, triu_ma, hicma.cut_intervals)

    else:
        under_noise = 0
        dist_list, chrom_list = hiCMatrix.getDistList(triu_ma.row, triu_ma.col, cut_intervals)

        assert len(dist_list) == len(triu_ma.data), "lists not of equal size"
        susprow_list = []
        suspcol_list = []
        transf_ma = np.zeros(len(triu_ma.data))
        start_time = time.time()
        # transform each value  of the data matrix to p-value, obs/exp, correlation etc.
        sys.stderr.write("computing transform values\n")
        for idx, data in enumerate(triu_ma.data):
            # skip if original value is less than noise level
            if noise_level:
                if per_chr:
                    if dist_list[idx] == -1:
                        continue
                    elif (orig_ma[triu_ma.row[idx],
                                  triu_ma.col[idx]] <=
                          noise_level[chrom_list[idx]]):
                        under_noise += 1
                        continue
                elif orig_ma[triu_ma.row[idx],
                             triu_ma.col[idx]] <= noise_level:
                    under_noise += 1
                    continue

            if method in ['nbinom-p-value', 'nbinom-expected']:
                if dist_list[idx] == -1:
                    continue
                if per_chr:
                    transf_ma[idx] = methods_avail[method](
                        data,
                        size[chrom_list[idx]][dist_list[idx]],
                        prob[chrom_list[idx]][dist_list[idx]])
                else:
                    transf_ma[idx] = methods_avail[method](data,
                                                           size[dist_list[idx]],
                                                           prob[dist_list[idx]])
                if data > 3 * orig_ma[triu_ma.row[idx], triu_ma.col[idx]]:
                    sys.stderr.write("skipping p-value {} for "
                                     "value {} at {}, norm-value {}\n".
                                     format(transf_ma[idx], chrom_list[idx],
                                            orig_ma[triu_ma.row[idx],
                                                    triu_ma.col[idx]], data))
                    continue
                if transf_ma[idx] > 4.5 and \
                        data > 2 * orig_ma[triu_ma.row[idx], triu_ma.col[idx]]:
                    susprow = triu_ma.row[idx]
                    suspcol = triu_ma.col[idx]
                    sys.stderr.write("suspicious p-value {} for "
                                     "value {} at {}, norm-value {}\n".
                                     format(transf_ma[idx], chrom_list[idx],
                                            orig_ma[susprow, suspcol], data))
                    susprow_list.append(susprow)
                    suspcol_list.append(suspcol)

            if method in ['obs/exp', 'residuals']:
                if per_chr:
                    if dist_list[idx] == -1:
                        continue
                    fit_mu = mu_[chrom_list[idx]]
                else:
                    fit_mu = mu_

                transf_ma[idx] = methods_avail[method](data,
                                                       fit_mu[dist_list[idx]])

            else:
                if per_chr:
                    if dist_list[idx] == -1:
                        continue
                    fit_mu = mu_[chrom_list[idx]]
                    fit_sigma = sigma[chrom_list[idx]]
                    fit_n = n_value[chrom_list[idx]]
                else:
                    fit_mu = mu_
                    fit_sigma = sigma
                    fit_n = n_value

                transf_ma[idx] = methods_avail[method](data,
                                                       fit_mu[dist_list[idx]],
                                                       fit_sigma[dist_list[idx]],
                                                       fit_n[dist_list[idx]])

            if idx > 0 and (idx == 10000 or idx % 500000 == 0):
                endtime = time.time()
                estimated = (float(len(transf_ma) - idx) *
                             (endtime - start_time)) / idx
                mmin, sec = divmod(estimated, 60)
                hour, mmin = divmod(mmin, 60)
                print "iteration: {} Estimated remaining time "\
                    "{:.0f}:{:.0f}:{:.0f}".format(idx, hour, mmin, sec)

        """
        print "problematic bins:"
        for uniq in np.concatenate([susprow_list, suspcol_list]):
            print hicma.cut_intervals[uniq]
        """
        # set the new values back into the original matrix
        triu_ma.data = transf_ma
        # fill the lower triangle
        triu_ma = triu_ma + scipy.sparse.triu(triu_ma, 1).T
        triu_ma = triu_ma.tocsr()
        triu_ma.eliminate_zeros()

    return triu_ma


def nbinom_est_dist(size, prob, triu_ma, cut_intervals):
    # compute a mapping from mean to distance
    mean2dist = {'mean': [], 'dist': []}
    for dist in np.sort(size.keys()):
        mean = scipy.stats.nbinom.mean(size[dist], prob[dist])
        if not np.isnan(mean):
            mean2dist['mean'].append(mean)
            mean2dist['dist'].append(dist)
    mean2dist['mean'] = np.array(mean2dist['mean'])
    mean2dist['dist'] = np.array(mean2dist['dist'])

    # the values have to be computed for all the
    # matrix excepting inter chromosome
    row, col = np.triu_indices(triu_ma.shape[0])
    dist_list, chrom_list = hiCMatrix.getDistList(row, col,
                                                  cut_intervals)
    triu_ma = triu_ma.tolil()
    transf_ma = np.zeros(len(dist_list))
    for idx, orig_dist in enumerate(dist_list):
        if orig_dist == -1:
            continue
        data = triu_ma[row[idx], col[idx]]
        try:
            size[orig_dist]
        except KeyError:
            continue
        if _nbinomPvalue(data,
                         size[orig_dist],
                         prob[orig_dist]) < 5:
            # pass
            continue

        # get largest closest mean
        if data > mean2dist['mean'].max() or \
                data < mean2dist['mean'].min():
            dist = orig_dist
        else:
            try:
                mean_idx = np.flatnonzero(data < mean2dist['mean'])[-1]
                dist = mean2dist['dist'][mean_idx]
            except IndexError:
                dist = orig_dist

        # min distance should be 1, otherwise the
        # sparse matrix will treat nans as cero distance
        transf_ma[idx] = dist_list[idx] - dist
        transf_ma[idx] = dist

    # set the new values back into the original matrix
    triu_ma = scipy.sparse.coo_matrix((transf_ma, (row, col)), shape=triu_ma.shape)
    # fill the lower triangle
    triu_ma = triu_ma + scipy.sparse.triu(triu_ma, 1).T
    triu_ma = triu_ma.tocsr()
    triu_ma.eliminate_zeros()

    return triu_ma


def applyFdr(matrix):
    """
    compute false discovery rate
    but only for half the symmetric matrix

    :param: matrix a sparse matrix
    """
    # use only the upper half of the matrix
    mat = scipy.sparse.triu(matrix)
    # exp is required because the matrix contains
    # -log(pvalues)
    mat.data = -np.log(_fdr(np.exp(-mat.data)))
    mat = mat + scipy.sparse.triu(mat, 1).T

    return mat


def _fdr(pvalues):
    """
    Code translated from the R function
    p.adjust(pvalues, method='fdr')

    :param: list of pvalues
    """
    len_pvals = len(pvalues)
    if len_pvals <= 1:
        return pvalues
    i = np.arange(1, len_pvals + 1)[::-1]
    order = np.argsort(pvalues)[::-1]
    rorder = np.argsort(order)

    # clip values bigger than 1 or smaller than 0
    return np.clip(float(len_pvals) / i * pvalues[order], 0, 1)[rorder]


def fitNegBinom_Rserve(countsByDistance, plot_distribution=False,
                       per_chr=False):
    """
    Fits a negative binomial distribution to the counts
    found at each different distance.

    The fitting is attempted first using a python method, and
    if this fails R is used through Rserve.

    For the fitting, the outliers are removed. Outliers are
    defined as those having a z-score higher than 3.4. This
    number as defined after exploring different values
    of z-scores and estimating the best goodness of fit.
    The HiC data is expected to contain outliers but they
    are problematic to fit and test the goodness of fit of a
    distribution, that's why there are removed.
    """

    # if the counts are per chromosome,
    # use the function recursively
    if per_chr:
        size = {}
        prob = {}
        for chrom in countsByDistance.keys():
            sys.stderr.write('computing negative binomial for '
                             '{}\n'.format(chrom))
            size[chrom], prob[chrom] = \
                fitNegBinom_Rserve(countsByDistance[chrom],
                                   plot_distribution=plot_distribution)
        return size, prob

    import pyRserve
    import matplotlib.pyplot as plt
    try:
        conn = pyRserve.connect()
        conn.r('library("MASS")')

    except:
        print "Could not connect to Rserve. Check that Rserve is up and running"
        exit(1)
    size = {}
    mu = {}
    prob = {}
    pval = {}
    good = 0
    bad = 0

    for dist in np.sort(countsByDistance.keys()):
        if dist == -1:  # skip intra chromosomal counts
            continue
        size[dist] = np.nan
        mu[dist] = np.nan
        prob[dist] = np.nan
        if sum(countsByDistance[dist]) == 0.0:
            print "no counts for bins at distance {}".format(dist)
            continue
        if np.any(np.isnan(countsByDistance[dist])) is True:
            exit("ERROR: matrix contains NaN values\n")

        counts = remove_outliers(countsByDistance[dist])
        if len(counts) <= 20:
            continue
        # the values in countsByDistance of a corrected matrix
        # are float values, but integers are needed for
        # the negative binomial fitting in R.
        counts_int = np.round(counts).astype('int')

        # try first using the python fit for the
        # negative binomial
        try:
            size[dist], prob[dist] = fit_nbinom(counts)
        except ValueError:
            # try with R..
            try:
                res = conn.r.fitdistr(counts_int, 'negative binomial')
            except:
                continue
            size[dist] = res[0]['size']
            mu[dist] = res[0]['mu']

            if np.isnan(size[dist]) or np.isnan(mu[dist]):
                sys.stderr.write("for dist={}, size={}, mu={}, "
                                 "len={}\n".format(dist, size[dist],
                                                   mu[dist], len(counts)))
                continue

            # The output from 'fitdistr' are size and mu.
            # but the scipy function that is based on the negative binomial
            # needs size and probability as parameters. However,
            # prob = size / ( size + mu )
            prob[dist] = size[dist] / (size[dist] + mu[dist])

        sys.stderr.write(".")  # print a . to show progress

        # evaluate fit of the counts distribution with respect to
        # the negative binomial  distribution using the parameters
        # returned by R
        fitted_dist = scipy.stats.nbinom.rvs(size[dist],
                                             prob[dist],
                                             size=len(counts) * 2)
        pval[dist] = scipy.stats.ks_2samp(counts_int,
                                          fitted_dist)[1]

#        pval[dist] = scipy.stats.wilcoxon(counts, fitted_dist)[1]
        if pval[dist] < 0.01:
            bad += 1
            sys.stderr.write(
                "\nThe fit p-value {} for {} is too low to consider "
                "the distribution negative binomial".format(
                    pval[dist], dist))
        else:
            good += 1

        if (plot_distribution and
                dist in [50000] + range(0, max(countsByDistance.keys()), 1000000)):
            # actual and fitted distributions are plotted
            # next to each other

            diff = counts.max() - counts.min()
            if diff >= 1000:
                nbins = 50
            elif 1000 > diff >= 100:
                nbins = 30
            elif 100 > diff >= 50:
                nbins = diff / 2
            else:
                nbins = (counts.max() - counts.min())
            freq, bins = np.histogram(counts.astype(int), nbins,
                                      normed=True)
            plt.hist(counts, bins, linewidth=0.1, alpha=0.8, normed=True)
            # plt.hist(fitted_dist, bins, histtype='step', linestyle='solid',
            #          linewidth=1.5, color='black', normed=True)
            pdf_fitted = scipy.stats.nbinom.pmf(bins.astype('int'),
                                                size[dist],
                                                prob[dist])
            plt.plot(bins.astype(int), pdf_fitted,
                     label='fitted nbinom gf={:.3f}'.format(pval[dist]))

            fig_name = '/tmp/fitt_{}_{}.png'.format('nbinom', dist)
            plt.title('{} bp; size: {}, prob: {}'.format(dist,
                                                         size[dist],
                                                         prob[dist]))
            plt.ylim(0, np.max(freq) + np.max(freq) * 0.2)
            plt.legend()
            plt.savefig(fig_name, dpi=200)
            plt.close()
            sys.stderr.write("check {}".format(fig_name))

    sys.stderr.write("good {}, bad {}\n".format(good, bad))

    return size, prob


def fitNegBinom(countsByDistance):
    """
    Replaced by fitNegBinom_Rserve
    Returns a tuple of two dictionaries,
    one contains the size fit and the
    other the probability
    """
    print "fit neg binom"
    size = {}
    mu = {}
    prob = {}
    pval = {}
    for dist in countsByDistance:
        size[dist] = np.nan
        mu[dist] = np.nan
        prob[dist] = np.nan
        if sum(countsByDistance[dist]) == 0.0:
            print "no counts for bins at distance {}".format(dist)
            continue
        if len(countsByDistance[dist]) <= 2:
            continue
        # the values in countsByDistance of a corrected matrix
        # are float values, but integers are needed for
        # the negative binomial.
        counts = np.rint(countsByDistance[dist]).astype('int')
        try:
            size[dist], prob[dist] = fit_nbinom(counts)
        except ValueError as error:
            print "could not compute pval for dist={}. "\
                "Message:\n {}".format(dist, error)

        if np.isnan(size[dist]) or np.isnan(prob[dist]):
            print "for dist={}, size={}, prob={}, len={}".format(dist,
                                                                 size[dist],
                                                                 prob[dist],
                                                                 len(counts))
        else:
            # evaluate fit of the counts distribution with respect to the
            # negative binomial distribution using the parameters returned by R
            pval[dist] = scipy.stats.wilcoxon(
                counts,
                scipy.stats.nbinom.rvs(size[dist],
                                       prob[dist],
                                       size=len(counts)))[1]
            if pval[dist] < 0.001:
                print "problem with {} when fitting a negative binomial. "\
                    "The fit p-value ({}) is too low to consider the "\
                    "distribution negative binomial".format(dist, pval[dist])

    return (size, prob)


def fitDistribution(countsByDistance, distribution, plot_distribution=False):
    """
    Generic method to fit continuous
    distributions to the  HiC countsByDistance
    The distribution names are the ones supported
    by scipy.
    """
    mu = {}
    sigma = {}
    pval = {}
    good = 0
    bad = 0
    good_nb = 0
    bad_nb = 0
    import pyRserve
    try:
        conn = pyRserve.connect()
        conn.r('library("MASS")')
    except:
        print "Could not connect to Rserve. Check that Rserve is up and running"
        exit(1)

    import sys
    for distnc in np.sort(countsByDistance.keys()):
        if distnc == -1:  # skip intra chromosomal counts
            continue
        if sum(countsByDistance[distnc]) == 0.0:
            print "no counts for bins at distance {}".format(distnc)
            continue
        if len(countsByDistance[distnc]) <= 2:
            continue
        sys.stderr.write('.')
        # TEMP code to compare with negative binomial ###

        # the values in countsByDistance of a corrected matrix
        # are float values, but integers are needed for
        # the negative binomial.

        counts_nb = remove_outliers(np.round(countsByDistance[distnc]).astype('int'))

        # try first using the python fit for the
        # negative binomial
        try:
            size, prob = fit_nbinom(remove_outliers(countsByDistance[distnc]))
        except ValueError:
            # try with R..
            res = conn.r.fitdistr(counts_nb, 'negative binomial')
            size = res[0]['size']
            mu_ = res[0]['mu']

            if np.isnan(size) or np.isnan(mu_):
                print "for dist={}, size={}, mu={}, len={}".format(
                    distnc, size, mu_, len(counts_nb))
                continue

            prob = size / (size + mu_)
        nbin = scipy.stats.nbinom(size, prob)
        #####

        counts = remove_outliers(countsByDistance[distnc])
        counts[counts == 0] = 0.01
        dist = getattr(scipy.stats, distribution)
        param = dist.fit(counts, floc=0)
        if np.any(np.isnan(param)):
            sys.stderr.write('\n{} no params computed'.format(distnc))
            import ipdb
            ipdb.set_trace()
        mu[distnc] = param[-1]
        sigma[distnc] = param[0]

        # estimate the goodness of fit pvalue
        fitted_dist = dist.rvs(*param[:-2],
                               loc=param[-2], scale=param[-1],
                               size=len(counts) * 2)
        pval[distnc] = scipy.stats.ks_2samp(counts,
                                            fitted_dist)[1]
        fitted_dist_nb = scipy.stats.nbinom.rvs(size,
                                                prob,
                                                size=len(counts_nb) * 2)

        pval_nb = scipy.stats.ks_2samp(counts_nb,
                                       fitted_dist_nb)[1]
        if pval[distnc] < 0.01:
            bad += 1
        else:
            good += 1

        if pval_nb < 0.01:
            bad_nb += 1
        else:
            good_nb += 1
        if pval[distnc] < 0.01:
            sys.stderr.write("\nproblem with {}, p-value for "
                             "{} fit: {} (NB fit: {})".format(distnc,
                                                              distribution,
                                                              pval[distnc],
                                                              pval_nb))

        if (plot_distribution and
                distnc in range(50000, max(countsByDistance.keys()), 500000)):

            import matplotlib.pyplot as plt
            freq, bins = np.histogram(counts, 30,
                                      normed=True)
            plt.close()  # to avoid overlaps
            plt.hist(counts, bins, linewidth=0.1, alpha=0.8, normed=True)
#            plt.hist(fitted_dist, bins, histtype='step', linestyle='solid',
#                      linewidth=1.5, color='black', normed=True)
#            plt.hist(fitted_dist_nb, bins, histtype='step', linestyle='solid',
#                      linewidth=1.5, color='grey', normed=True)
            pdf_fitted = dist.pdf(bins,
                                  *param[:-2],
                                  loc=param[-2],
                                  scale=param[-1])
            ##
            plt.plot(bins.astype(int), nbin.pmf(bins.astype('int')),
                     label='NB {:.2f}'.format(pval_nb))
            ##
            plt.plot(bins, pdf_fitted,
                     label='{} {:.2f}'.format(distribution, pval[distnc]))
            fig_name = '/tmp/fitt_{}_{}.png'.format(distribution, distnc)
            plt.title('{} bp'.format(distnc))
            plt.ylim(0, np.max(freq) + np.max(freq) * 0.2)
            plt.legend()
            plt.savefig(fig_name, dpi=200)
            plt.close()
            print "check {}".format(fig_name)
    print "good {}, bad {}, good_nb {}, bad_nb {}".format(good, bad, good_nb,
                                                          bad_nb)
    return (mu, sigma)


def remove_outliers(data, max_deviation=3.5):
    """
    The method is based on the median absolute deviation. See
    Boris Iglewicz and David Hoaglin (1993),
    "Volume 16: How to Detect and Handle Outliers",
    The ASQC Basic References in Quality Control:
    Statistical Techniques, Edward F. Mykytka, Ph.D., Editor.

    returns the list, without the outliers
    """
    median = np.median(data)
    b_value = 1.4826  # value for normal distribution
    mad = b_value * np.median(np.abs(data - median))

    if mad > 0:
        deviation = abs(data - median) / mad
        """
        outliers = data[deviation > max_deviation]
        print "outliers removed {}".format(len(outliers))
        print outliers
        """
        data = data[deviation <= max_deviation]
    return data


def fitChisquared(countsByDistance):
    shape = {}
    loc = {}
    scale = {}
    pval = {}
    for x in countsByDistance:
        if len(countsByDistance[x]) > 2:
            counts = countsByDistance[x]
            if len(counts) > 10000:
                rand = np.arange(10000)
                np.random.shuffle(rand)
                counts = counts[rand]
            shape[x], loc[x], scale[x] = scipy.stats.chi2.fit(counts, 3, floc=0)
            # estimate the fit pvalue
            pval[x] = scipy.stats.wilcoxon(counts,
                                           scipy.stats.chi2.rvs(shape[x],
                                                                loc=loc[x],
                                                                scale=scale[x],
                                                                size=len(counts)))[1]
            if pval[x] < 0.001:
                print "problem with {}, p-value for log-norm fit: {}".format(x, pval[x])

    return (shape, loc, scale)


def _residuals(value, mu):
    return value - mu


def _obsExp(value, mu):
    return value / mu


def _zscore(value, mu, sigma, n):
    return (value - mu) / sigma


def _tscore(value, mu, sigma, n):
    return (value - mu) / (sigma / np.sqrt(n))


def _pvalue(value, mu, sigma, n):
    #    tscore = _tscore(value, mu, sigma, n)
    pvalue = -np.log(scipy.stats.t.sf(value, n, loc=mu, scale=sigma))
    return pvalue


def _lognormPvalue(value, mu, sigma, n):
    """
    n is not used, but is generic
    to other calls (see _pvalue).
    """
    return scipy.stats.norm.sf(np.log(value), mu, sigma)


def _chi2Pvalue(value, shape, loc, scale):
    """
    n is not used, but is generic
    to other calls (see _pvalue).
    """
    return -np.log(scipy.stats.chi2.sf(np.log(value), shape, loc=loc, scale=scale))


def _nbinomPvalue(value, size, prob):
    pvalue = -np.log(scipy.stats.nbinom.sf(value, size, prob))
    return pvalue


def _nbinomExpected(value, size, prob):
    mean = scipy.stats.nbinom.mean(size, prob)
    return mean


def convertNansToZeros(ma):
    ma.data[np.flatnonzero(np.isnan(ma.data))] = 0
    # ma[np.where(np.isnan(ma) == True)] = 0
    return ma


def myAverage(valuesArray, avgType='mean'):

    valuesArray = valuesArray[np.logical_not(np.isnan(valuesArray))]
    if avgType == 'mean':
        mean = np.mean(valuesArray)
    else:
        mean = np.median(valuesArray)

    return mean


def enlarge_bins(bin_intervals):
    r"""
    takes a list of consecutive, but not
    directly touching, bin intervals
    and joins them such that the
    end and start of consecutive bins
    is the same.

    >>> bin_intervals = [('chr1', 10, 30, 1), ('chr1', 50, 80, 2),
    ... ('chr2', 10, 60, 3), ('chr2', 60, 90, 4)]
    >>> enlarge_bins(bin_intervals)
    [('chr1', 0, 40, 1), ('chr1', 40, 80, 2), ('chr2', 0, 60, 3), ('chr2', 60, 90, 4)]
    """
    # enlarge remaining bins
    chr_start = True
    for idx in range(len(bin_intervals) - 1):
        chrom, start, end, extra = bin_intervals[idx]
        chrom_next, start_next, end_next, extra_next = bin_intervals[idx + 1]
        if chr_start is True:
            start = 0
            chr_start = False
        if chrom == chrom_next and \
                end != start_next:
            middle = start_next - int((start_next - end) / 2)
            bin_intervals[idx] = (chrom, start, middle, extra)
            bin_intervals[idx + 1] = (chrom, middle, end_next, extra_next)
        if chrom != chrom_next:
            bin_intervals[idx] = (chrom, start, end, extra)
            bin_intervals[idx + 1] = (chrom_next, 0, end_next, extra_next)

    chrom, start, end, extra = bin_intervals[-1]
    bin_intervals[-1] = (chrom, start, end, extra)

    return bin_intervals


def genomicRegion(string):
    """
    validates and cleans up a string corresponding to a genomic
    region in the form ideally of chromosome:start-end
    but other forms are also possible like start
    and end containing comas.

    This code is intended to be used to validate and
    format a argparse parameter.

    :return: string in the form chrom:start:end
    """
    # remove whitespaces using split,join trick
    region = ''.join(string.split())
    if region == '':
        return None
    # remove undesired characters that may be present and
    # replace - by :
    region = region.translate(None, ",;|!{}()").replace("-", ":")
    if len(region) == 0:
        raise argparse.ArgumentTypeError(
            "{} is not a valid region".format(string))
    return region


def getUserRegion(chromSizes, regionString, max_chunk_size=1e6):
    """
    Verifies if a given region argument, given by the user
    is valid. The format of the regionString is chrom:start:end:tileSize
    where start, end and tileSize are optional.

    # this should work in doctest but it does not. So I
    # commented it.
    #>>> data = getUserRegion({'chr2': 1000}, "chr1:10:10")
    #Traceback (most recent call last):
    #    ...
    #NameError: Unknown chromosome: chr1
    #Known chromosomes are: ['chr2']

    >>> getUserRegion({'chr2': 1000}, "chr2:10:1001")
    ([('chr2', 1000)], 10, 1000, 990)

    Test chunk and regions size reduction to match tile size
    >>> getUserRegion({'chr2': 200000}, "chr2:10:123344:3")
    ([('chr2', 123344)], 9, 123345, 123336)
    """
    region = regionString.split(":")
    chrom = region[0]
    chromSizes = dict(chromSizes)

    try:
        chromSizes[chrom]
    except KeyError:
        raise NameError("Unknown chromosome: %s\nKnown "
                        "chromosomes are: %s " % (chrom, chromSizes.keys()))
    try:
        regionStart = int(region[1])
    except IndexError:
        regionStart = 0
    try:
        regionEnd = int(region[2]) if int(region[2]) <= chromSizes[chrom] \
            else chromSizes[chrom]
    except IndexError:
        regionEnd = chromSizes[chrom]
    if regionStart > regionEnd or regionStart < 0:
        raise NameError("%s not valid. The format is chrom:start:end. "
                        "Without comas, dashes or dots. " % (regionString))
    try:
        tilesize = int(region[3])
    except IndexError:
        tilesize = None

    chromSizes = [(chrom, regionEnd)]

    # if tilesize is given, make regionStart and regionEnd
    # multiple of tileSize
    if tilesize:
        regionStart -= regionStart % tilesize
        regionEnd += tilesize - (regionEnd % tilesize)

    chunkSize = int(regionEnd - regionStart)
    if chunkSize > max_chunk_size:
        chunkSize = max_chunk_size
        if tilesize and tilesize < chunkSize:
            chunkSize -= chunkSize % tilesize

    return (chromSizes, regionStart, regionEnd, int(chunkSize))


def fit_nbinom(k):
    """
    from https://gist.github.com/gjx/5987413
    Note: for some cases this function does
    not work. brentq fails with
    error: f(a) and f(b) must have different signs
    Rserve may be more accurate
    """
    from scipy.special import psi
    from scipy.optimize import brentq
    N = len(k)
    n = brentq(lambda r: sum(psi(k + r)) - N * psi(r) +
               N * np.log(r / (r + sum(k / N))),
               np.finfo(np.float128).eps,
               np.max(k))
    p = n / (n + sum(k / N))  # Note: this `p` = 1 - `p` from Wikipedia
    return n, p



import argparse
import sys
import numpy as np
from scipy.sparse import coo_matrix, dia_matrix, dok_matrix
import time
from os import unlink
import os
import pysam
from intervaltree import IntervalTree, Interval

# own tools
from hicexplorer import HiCMatrix as hm
from hicexplorer.utilities import getUserRegion, genomicRegion
from hicexplorer._version import __version__
import hicexplorer.hicPrepareQCreport as QC

debug = 1


class ReadPositionMatrix(object):
    """ class to check for PCR duplicates.
    A sparse matrix having as bins all possible
    start sites (single bp resolution)
    is created. PCR duplicates
    are determined by checking if the matrix
    cell is already filled.

    """

    def __init__(self, chrom_sizes):
        """
        >>> rp = ReadPositionMatrix([('1', 10), ('2', 10)])
        >>> rp.pos2matrix_bin('1', 0)
        0
        >>> rp.pos2matrix_bin('2', 0)
        10
        >>> rp.is_duplicated('1', 0, '2', 0)
        False
        >>> rp.pos_matrix[0,10]
        True
        >>> rp.pos_matrix[10,0]
        True
        >>> rp.is_duplicated('1', 0, '2', 0)
        True
        """
        # determine number of bins
        total_size = 0
        self.chr_start_pos = {}
        for chrom, size in chrom_sizes:
            self.chr_start_pos[chrom] = total_size
            total_size += size

        self.pos_matrix = dok_matrix((total_size, total_size), dtype=bool)

    def is_duplicated(self, chrom1, start1, chrom2, start2):
        pos1 = self.pos2matrix_bin(chrom1, start1)
        pos2 = self.pos2matrix_bin(chrom2, start2)
        if self.pos_matrix[pos1, pos2] == 1:
            return True
        else:
            self.pos_matrix[pos1, pos2] = 1
            self.pos_matrix[pos2, pos1] = 1
            return False

    def pos2matrix_bin(self, chrom, start):
        return self.chr_start_pos[chrom] + start


def parse_arguments(args=None):

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=('Using an alignment from Bowtie2 where both '
                     'PE reads are mapped using  the --local '
                     'option, this program reads such file and '
                     'creates a matrix of interactions.'))

    # define the arguments
    parser.add_argument('--samFiles', '-s',
                        help='The two sam files to process',
                        metavar='two sam files',
                        nargs=2,
                        type=argparse.FileType('r'),
                        required=True)

    # define the arguments
    parser.add_argument('--outBam', '-b',
                        help='Bam file to process',
                        metavar='bam file',
                        type=argparse.FileType('w'),
                        required=True)

    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument('--binSize', '-bs',
                       help='Size in bp for the bins.',
                       type=int,
                       default=10000)

    group.add_argument('--restrictionCutFile', '-rs',
                       help=('BED file with all restriction cut places. '
                             'Should contain only  mappable '
                             'restriction sites. If given, the bins are '
                             'set to match the restriction fragments (i.e. '
                             'the region between one restriction site and '
                             'the next).'),
                       type=argparse.FileType('r'),
                       metavar='BED file')

    parser.add_argument('--minDistance',
                        help='Minimum distance between restriction sites. '
                        'Restriction sites that are closer that this '
                        'distance are merged into one. This option only '
                        'applies if --restrictionCutFile is given.',
                        type=int,
                        default=300,
                        required=False)

    parser.add_argument('--maxDistance',
                        help='Maximum distance in bp from restriction site '
                        'to read, to consider a read a valid one. This option '
                        'only applies if --restrictionCutFile is given.',
                        type=int,
                        default=800,
                        required=False)

    parser.add_argument('--restrictionSequence', '-seq',
                        help='Sequence of the restriction site. This is used '
                        'to discard reads that end/start with such sequence '
                        'and that are considered un-ligated fragments or '
                        '"dangling-ends". If not given, such statistics will '
                        'not be available.')

    parser.add_argument('--outFileName', '-o',
                        help='Output file name for a matrix',
                        metavar='FILENAME',
                        type=argparse.FileType('w'),
                        required=True)

    parser.add_argument('--QCfolder',
                        help='Path of folder to save the quality control data for the matrix',
                        metavar='FOLDER',
                        required=True)

    parser.add_argument('--region', '-r',
                        help='Region of the genome to limit the operation. '
                        'The format is chr:start-end. Also valid is just to '
                        'specify a chromosome, for example --region chr10',
                        metavar="CHR:START-END",
                        required=False,
                        type=genomicRegion
                        )

    parser.add_argument('--removeSelfLigation',
                        help=argparse.SUPPRESS,
                        required=False,
                        default=True
                        )

    parser.add_argument('--removeSelfCircles',
                        help='If set, outward facing reads, at a distance of less thatn 25kbs are removed.',
                        required=False,
                        action='store_true'
                        )

    parser.add_argument('--minMappingQuality',
                        help='minimun mapping quality for reads to be accepted. Because the restriction '
                             'enzyme site could be located on top of the read, this may reduce the '
                             'reported quality of the read. Thus, this parameter may be adusted if too many '
                             'low quality (but otherwise perfectly valid hic-reads) are found. A good strategy '
                             'is to make a test run (using the --doTestRun), then checking the results to see '
                             'if too many low quality reads are present and then using the bam file generated to '
                             'check if those low quality reads are caused by the read not being mapped entirely.',
                        required=False,
                        default=15,
                        type=int
                        )

    parser.add_argument('--doTestRun',
                        help='A test run is useful to test the quality of a Hi-C experiment quickly. It works by '
                             'testing only 1,000.000 reads. This option is useful to get an idea of quality control'
                             'values like inter-chromosomal interactins, duplication rates etc.',
                        action='store_true'
                        )

    parser.add_argument('--skipDuplicationCheck',
                        help='Identification of duplicated read pairs is memory consuming. Thus, in case of '
                             'memory errors this check can be skipped. However, consider running a `--doTestRun` '
                             'first to get an estimation of the duplicated reads. ',
                        action='store_true'
                        )

    parser.add_argument('--version', action='version',
                        version='%(prog)s {}'.format(__version__))

    return parser


def intervalListToIntervalTree(interval_list):
    """
    given a dictionary containing tuples of chrom, start, end,
    this is transformed to an interval trees. To each
    interval an id is assigned, this id corresponds to the
    position of the interval in the given array of tuples
    and if needed can be used to identify
    the index of a row/colum in the hic matrix.

    >>> bin_list = [('chrX', 0, 50000), ('chrX', 50000, 100000)]
    >>> res = intervalListToIntervalTree(bin_list)
    >>> sorted(res['chrX'])
    [Interval(0, 50000, 0), Interval(50000, 100000, 1)]
    """
    bin_int_tree = {}

    for intval_id, intval in enumerate(interval_list):
        chrom, start, end = intval[0:3]
        if chrom not in bin_int_tree:
            bin_int_tree[chrom] = IntervalTree()
        bin_int_tree[chrom].add(Interval(start, end, intval_id))

    return bin_int_tree


def get_bins(bin_size, chrom_size, region=None):
    """
    Split the chromosomes into even sized bins
    of length bin_size.

    Returns a list of tuples containing
    ('chrom', start, end)

    >>> test = Tester()
    >>> chrom_size = get_chrom_sizes(pysam.Samfile(test.bam_file_1))
    >>> get_bins(50000, chrom_size)
    [('contig-2', 0, 3345), ('contig-1', 0, 7125)]
    >>> get_bins(50000, chrom_size, region='contig-1')
    [('contig-1', 0, 7125)]
    """
    bin_intvals = []
    start = 0
    if region:
        chrom_size, start, _, _ = \
            getUserRegion(chrom_size, region)

    for chrom, size in chrom_size:
        for interval in xrange(start, size, bin_size):
            bin_intvals.append((chrom, interval,
                                min(size, interval + bin_size)))
    return bin_intvals


def bed2interval_list(bed_file_handler):
    r"""
    reads a BED file and returns
    a list of tuples containing
    (chromosome name, start, end)

    >>> import tempfile, os

    Make a temporary BED file
    >>> _file = tempfile.NamedTemporaryFile(delete=False)
    >>> _file.write('chr1\t10\t20\tH1\t0\n')
    >>> _file.write("chr1\t60\t70\tH2\t0\n")
    >>> _file.close()
    >>> bed2interval_list(open(_file.name))
    [('chr1', 10, 20), ('chr1', 60, 70)]
    >>> os.remove(_file.name)
    """
    count = 0
    interval_list = []
    for line in bed_file_handler:
        count += 1
        fields = line.strip().split()
        try:
            chrom, start, end = fields[0], int(fields[1]), int(fields[2])
        except IndexError:
            sys.stderr.write("error reading BED file at line {}".format(count))

        interval_list.append((chrom, start, end))
    return interval_list


def get_rf_bins(rf_cut_intervals, min_distance=200, max_distance=800):
    r"""
    returns a list of tuples containing a bin having the format:
    ('chrom', start, end)

    The goal is to have bins containing each  a restriction site close to
    the center. Restriction sites that are less than 'min_distance' appart
    are merged. To the left and right of the restriction site
    'max_distance' bp are added unless they clash with other bin.
    Resulting bins are not immediately one after the other.

    The given cut sites list has to be ordered chr, and start site

    :param rf_cut_intervals: list of tuples containing the position of restriction fragment sites
    :param min_distance: min distance between restriction fragment sites.
    :param max_distance: max distance between restriction fragment and bin border.
    :return:

    # two restriction sites of length 10
    >>> rf_cut_interval = [('chr1', 10, 20), ('chr1', 60, 70)]

    The following two bins are close together
    and should be  merged under min_distance = 20
    >>> rf_cut_interval.extend([('chr2', 20, 30), ('chr2', 40, 50),
    ... ('chr2', 70, 80)])
    >>> get_rf_bins(rf_cut_interval, min_distance=10, max_distance=20)
    [('chr1', 0, 40), ('chr1', 40, 90), ('chr2', 0, 60), ('chr2', 60, 100)]
    """
    sys.stderr.write("Minimum distance considered between "
                     "restriction sites is {}\nMax "
                     "distance: {}\n".format(min_distance, max_distance))

    chrom, start, end = zip(*rf_cut_intervals)
    rest_site_len = end[0] - start[0]

    # find sites that are less than min_distance apart
    to_merge = np.flatnonzero(np.diff(start) - rest_site_len <= min_distance)
    to_merge += 1  # + 1 to account for np.diff index handling
    merge_idx = 0
    # add max_distance to both sides
    start = np.array(start) - max_distance
    end = np.array(end) + max_distance
    new_start = [max(0, start[0])]
    new_end = []
    new_chrom = [chrom[0]]
    for idx in range(1, len(start)):
        # identify end of chromosome
        if chrom[idx] != chrom[idx - 1]:
            new_start.append(max(0, start[idx]))
            new_end.append(end[idx - 1])
            new_chrom.append(chrom[idx])
            merge_idx += 1
            continue

        if merge_idx < len(to_merge) and idx == to_merge[merge_idx]:
            merge_idx += 1
            continue

        # identify overlapping bins
        if chrom[idx] == chrom[idx - 1] and end[idx - 1] > start[idx]:
            middle = start[idx] + int((end[idx - 1] - start[idx]) / 2)
            new_start.append(middle)
            new_end.append(middle)
        else:
            new_start.append(start[idx])
            new_end.append(end[idx - 1])
        new_chrom.append(chrom[idx])

    new_end.append(end[-1])
    assert len(new_chrom) == len(new_start), "error"
    assert len(new_end) == len(new_start), "error"

    intervals = zip(new_chrom, new_start, new_end)
    intervals = [(_chrom, _start, _end) for _chrom, _start, _end in intervals if _end - _start >= min_distance]
    return intervals


def get_chrom_sizes(bam_handle):
    """
    return the list of chromosome names and their
    size from the bam file
    The return value is a list of the form
    [('chr1', 2343434), ('chr2', 43432432)]

    >>> test = Tester()
    >>> get_chrom_sizes(pysam.Samfile(test.bam_file_1, 'rb'))
    [('contig-2', 3345), ('contig-1', 7125)]
    """

    # in some cases there are repeated entries in
    # the bam file. Thus, I first convert to dict,
    # then to list.
    list_chrom_sizes = dict(zip(bam_handle.references,
                                bam_handle.lengths))
    return list_chrom_sizes.items()


def check_dangling_end(read, dangling_sequences):
    """
    given a pysam read object, this function
    checks if a forward read starts with
    the dangling sequence or if a reverse
    read ends with the dangling sequence.
    """
    ds = dangling_sequences
    # skip forward read that stars with the restriction sequence
    if not read.is_reverse and \
            read.seq.upper()[0:len(ds['pat_forw'])] == ds['pat_forw']:
        return True

    # skip reverse read that ends with the restriction sequence
    if read.is_reverse and \
            read.seq.upper()[-len(ds['pat_rev']):] == ds['pat_rev']:
        return True

    return False


def get_supplementary_alignment(read, pysam_obj):
    """Checks if a read has a supplementary alignment
    :param read pysam AlignedSegment
    :param pysam_obj pysam file object

    :return pysam AlignedSegment of the supplementary aligment or None in case of no supplementary alignment
    """

    # the SA field contains a list of other alignments as a ';' delimited list in the format
    # rname,pos,strand,CIGAR,mapQ,NM;
    if read.has_tag('SA'):
        # field always ends in ';' thus last element after split is always empty, hence [0:-1]
        other_alignments = read.get_tag('SA').split(";")[0:-1]
        supplementary_alignment = []
        for i in range(len(other_alignments)):
            _sup = pysam_obj.next()
            if _sup.is_supplementary and _sup.qname == read.qname:
                supplementary_alignment.append(_sup)

        return supplementary_alignment

    else:
        return None


def get_correct_map(primary, supplement_list):
    """
    Decides which of the mappings, the primary or supplement, is correct. In the case of
    long reads (eg. 150bp) the restriction enzyme site could split the read into two parts
    but only the mapping corresponding to the start of the read should be considered as
    the correct one.

    For example:

    Forward read:

                          _  <- cut site
    |======================|==========>
                           -
    |----------------------|
        correct mapping


    reverse read:

                          _  <- cut site
    <======================|==========|
                           -
                           |----------|
                           correct mapping


    :param primary: pysam AlignedSegment for primary mapping
    :param supplement_list: list of pysam AlignedSegment for secondary mapping

    :return: pysam AlignedSegment that is mapped correctly
    """

    for supplement in supplement_list:
        assert primary.qname == supplement.qname, "ERROR, primary " \
            "and supplementary reads do not have the same id. The ids " \
            "are as follows\n{}\n{}".format(primary.qname, supplement.qname)
    read_list = [primary] + supplement_list
    first_mapped = []
    for idx, read in enumerate(read_list):
        if read.is_reverse:
            cigartuples = read.cigartuples[::-1]
        else:
            cigartuples = read.cigartuples[:]

        first_mapped.append([x for x, cig in enumerate(cigartuples) if cig[0] == 0][0])
    # find which read has a cigar string that maps first than any of the others.
    idx_min = first_mapped.index(min(first_mapped))

    return read_list[idx_min]


def enlarge_bins(bin_intervals, chrom_sizes):
    r"""
    takes a list of consecutive but not
    one after the other bin intervals
    and joins them such that the
    end and start of consecutive bins
    is the same.

    >>> chrom_sizes = [('chr1', 100), ('chr2', 100)]
    >>> bin_intervals =     [('chr1', 10, 30), ('chr1', 50, 80),
    ... ('chr2', 10, 60), ('chr2', 60, 90)]
    >>> enlarge_bins(bin_intervals, chrom_sizes)
    [('chr1', 0, 40), ('chr1', 40, 100), ('chr2', 0, 60), ('chr2', 60, 100)]
    """
    # enlarge remaining bins
    chr_start = True
    chrom_sizes_dict = dict(chrom_sizes)
    for idx in range(len(bin_intervals) - 1):
        chrom, start, end = bin_intervals[idx]
        chrom_next, start_next, end_next = bin_intervals[idx + 1]
        if chr_start is True:
            start = 0
            chr_start = False
        if chrom == chrom_next and \
                end != start_next:
            middle = start_next - (start_next - end) / 2
            bin_intervals[idx] = (chrom, start, middle)
            bin_intervals[idx + 1] = (chrom, middle, end_next)
        if chrom != chrom_next:
            bin_intervals[idx] = (chrom, start, chrom_sizes_dict[chrom])
            bin_intervals[idx + 1] = (chrom_next, 0, end_next)

    chrom, start, end = bin_intervals[-1]
    bin_intervals[-1] = (chrom, start, chrom_sizes_dict[chrom])

    return bin_intervals


def main(args=None):
    """
    Reads line by line two bam files that are not sorted.
    Each line in the two bam files should correspond
    to the mapped position of the two ends of a Hi-C
    fragment.

    Each mate pair is assessed to determine if it is
    a valid Hi-C pair, in such case a matrix
    reporting the counts of mates is constructed.

    A bam file containing the valid Hi-C reads
    is also constructed
    """

    # check pysam version
    from distutils.version import LooseVersion
    if LooseVersion(pysam.__version__) < LooseVersion("0.8.3"):
        exit("\n*ERROR*\n\nVersion of pysam has to be higher than 0.8.3. Current installed version is {}\n".format(pysam.__version__))

    args = parse_arguments().parse_args(args)

    # check that the log folder is valid
    try:
        QC.make_sure_path_exists(args.QCfolder)
    except OSError:
        exit("Can't open/create QC folder path: {}. Please check".format(args.QCfolder))

    sys.stderr.write("reading {} and {} to build hic_matrix\n".format(args.samFiles[0].name,
                                                                      args.samFiles[1].name))
    str1 = pysam.Samfile(args.samFiles[0].name, 'rb')
    str2 = pysam.Samfile(args.samFiles[1].name, 'rb')

    args.samFiles[0].close()
    args.samFiles[1].close()
    args.outBam.close()
    out_bam = pysam.Samfile(args.outBam.name, 'wb', template=str1)

    chrom_sizes = get_chrom_sizes(str1)
    # initialize read start positions matrix
    read_pos_matrix = ReadPositionMatrix(chrom_sizes)

    # define bins
    rf_positions = None
    if args.restrictionCutFile:
        rf_interval = bed2interval_list(args.restrictionCutFile)
        bin_intervals = get_rf_bins(rf_interval,
                                    min_distance=args.minDistance,
                                    max_distance=args.maxDistance)

        rf_positions = intervalListToIntervalTree(rf_interval)
    else:
        bin_intervals = get_bins(args.binSize, chrom_sizes, args.region)

    sys.stderr.write("Matrix size: {}\n".format(len(bin_intervals)))
    matrix_size = len(bin_intervals)
    bin_intval_tree = intervalListToIntervalTree(bin_intervals)
    ref_id2name = str1.references

    dangling_sequences = dict()
    if args.restrictionSequence:
        # build a list of dangling sequences
        args.restrictionSequence = args.restrictionSequence.upper()
        dangling_sequences['pat_forw'] = args.restrictionSequence[1:]
        dangling_sequences['pat_rev'] = args.restrictionSequence[:-1]
        sys.stderr.write("dangling sequences to check "
                         "are {}\n".format(dangling_sequences))

    # initialize coverage vectors that
    # save the number of reads that overlap
    # a bin.
    # To save memory, coverage is not measured by bp
    # but by bins of length 10bp
    coverage = []
    binsize = 10
    for value in bin_intervals:
        chrom, start, end = value
        coverage.append(np.zeros((end - start) / binsize, dtype='int'))

    start_time = time.time()
    pair_added = 0
    one_mate_unmapped = 0
    one_mate_low_quality = 0
    one_mate_not_unique = 0
    dangling_end = 0
    self_circle = 0
    self_ligation = 0
    same_fragment = 0
    mate_not_close_to_rf = 0
    duplicated_pairs = 0

    count_inward = 0
    count_outward = 0
    count_left = 0
    count_right = 0
    inter_chromosomal = 0
    short_range = 0
    long_range = 0

    iter_num = 0
    row = []
    col = []
    data = []
    hic_matrix = None
    # read the sam files line by line

    while True:
        iter_num += 1
        if iter_num % 1e6 == 0:
            elapsed_time = time.time() - start_time
            sys.stderr.write("processing {} lines took {:.2f} "
                             "secs ({:.1f} lines per "
                             "second)\n".format(iter_num,
                                                elapsed_time,
                                                iter_num / elapsed_time))
            sys.stderr.write("{} ({:.2f}%) valid pairs added to matrix"
                             "\n".format(pair_added, float(100 * pair_added) / iter_num))
        if args.doTestRun and iter_num > 1e5:
            sys.stderr.write("\n## *WARNING*. Early exit because of --doTestRun parameter  ##\n\n")
            break
        try:
            mate1 = str1.next()
            mate2 = str2.next()
        except StopIteration:
            break

        # skip 'not primary' alignments
        while mate1.flag & 256 == 256:
            mate1 = str1.next()

        while mate2.flag & 256 == 256:
            mate2 = str2.next()

        assert mate1.qname == mate2.qname, "FATAL ERROR {} {} " \
            "Be sure that the sam files have the same read order " \
            "If using Bowtie2 or Hisat2 add " \
            "the --reorder option".format(mate1.qname, mate2.qname)

        # check for supplementary alignments
        # (needs to be done before skipping any unmapped reads
        # to keep the order of the two bam files in sync)
        mate1_supplementary_list = get_supplementary_alignment(mate1, str1)
        mate2_supplementary_list = get_supplementary_alignment(mate2, str2)

        if mate1_supplementary_list:
            mate1 = get_correct_map(mate1, mate1_supplementary_list)

        if mate2_supplementary_list:
            mate2 = get_correct_map(mate2, mate2_supplementary_list)

        # skip if any of the reads is not mapped
        if mate1.flag & 0x4 == 4 or mate2.flag & 0x4 == 4:
            one_mate_unmapped += 1
            continue

        # skip if the read quality is low
        if mate1.mapq < args.minMappingQuality or mate2.mapq < args.minMappingQuality:
            # for bwa other way to test
            # for multi-mapping reads is with a mapq = 0
            # the XS flag is not reliable.
            if mate1.mapq == 0 & mate2.mapq == 0:
                one_mate_not_unique += 1
                continue

            """
            # check if low quality is because of
            # read being repetitive
            # by reading the XS flag.
            # The XS:i field is set by bowtie when a read is
            # multi read and it contains the mapping score of the next
            # best match
            if 'XS' in dict(mate1.tags) or 'XS' in dict(mate2.tags):
                one_mate_not_unique += 1
                continue
            """

            one_mate_low_quality += 1
            continue

        if args.skipDuplicationCheck is False:
            if read_pos_matrix.is_duplicated(ref_id2name[mate1.rname],
                                             mate1.pos,
                                             ref_id2name[mate2.rname],
                                             mate2.pos):
                duplicated_pairs += 1
                continue

        # check if reads belong to a bin
        mate_bins = []
        mate_is_unasigned = False
        for mate in [mate1, mate2]:
            mate_ref = ref_id2name[mate.rname]
            # find the middle genomic position of the read. This is used to find the bin it belongs to.
            read_middle = mate.pos + int(mate.qlen / 2)
            try:
                mate_bin = sorted(bin_intval_tree[mate_ref][read_middle:read_middle + 1])
            except KeyError:
                # for small contigs it can happen that they are not
                # in the bin_intval_tree keys if no restriction site is found on the contig.
                mate_is_unasigned = True
                break

            # report no match case
            if len(mate_bin) == 0:
                mate_is_unasigned = True
                break
            # take by default only the first match
            # (although always there should be only
            # one match
            mate_bin = mate_bin[0]

            mate_bin_id = mate_bin.data
            mate_bins.append(mate_bin_id)

        # if a mate is unassigned, it means it is not close
        # to a restriction sites
        if mate_is_unasigned is True:
            mate_not_close_to_rf += 1
            continue

        # check if mates are in the same chromosome
        if mate1.reference_id != mate2.reference_id:
            orientation = 'diff_chromosome'
        else:
            # to identify 'inward' and 'outward' orientations
            # the order or the mates in the genome has to be
            # known.
            if mate1.pos < mate2.pos:
                first_mate = mate1
                second_mate = mate2
            else:
                first_mate = mate2
                second_mate = mate1

            """
            outward
            <---------------              ---------------->

            inward
            --------------->              <----------------

            same-strand-right
            --------------->              ---------------->

            same-strand-left
            <---------------              <----------------
            """

            if not first_mate.is_reverse and second_mate.is_reverse:
                orientation = 'inward'
            elif first_mate.is_reverse and not second_mate.is_reverse:
                orientation = 'outward'
            elif first_mate.is_reverse and second_mate.is_reverse:
                orientation = 'same-strand-left'
            else:
                orientation = 'same-strand-right'

            # check self-circles
            # self circles are defined as pairs within 25kb
            # with 'outward' orientation (Jin et al. 2013. Nature)
            if abs(mate2.pos - mate1.pos) < 25000 and orientation == 'outward':
                self_circle += 1
                if args.removeSelfCircles:
                    continue

            # check for dangling ends if the restriction sequence
            # is known:
            if args.restrictionSequence:
                if check_dangling_end(mate1, dangling_sequences) or \
                        check_dangling_end(mate2, dangling_sequences):
                    dangling_end += 1
                    continue

            if abs(mate2.pos - mate1.pos) < 1000 and orientation == 'inward':
                has_rf = []

                if rf_positions and args.restrictionSequence:
                    # check if in between the two mate
                    # ends the restriction fragment is found.

                    # the interval used is:
                    # start of fragment + length of restriction sequence
                    # end of fragment - length of restriction sequence
                    # the restriction sequence length is subtracted
                    # such that only fragments internally containing
                    # the restriction site are identified
                    frag_start = min(mate1.pos, mate2.pos) + len(args.restrictionSequence)
                    frag_end = max(mate1.pos + mate1.qlen, mate2.pos + mate2.qlen) - len(args.restrictionSequence)
                    mate_ref = ref_id2name[mate1.rname]
                    has_rf = sorted(rf_positions[mate_ref][frag_start: frag_end])

                # case when there is no restriction fragment site between the mates
                if len(has_rf) == 0:
                    same_fragment += 1
                    continue

                self_ligation += 1
                if args.removeSelfLigation:
                    # skip self ligations
                    continue

            # set insert size to save bam
            mate1.isize = mate2.pos - mate1.pos
            mate2.isize = mate1.pos - mate2.pos

        # if mate_bins, which is set in the previous section
        # does not have size=2, it means that one
        # of the exceptions happened
        if len(mate_bins) != 2:
            continue

        # count type of pair (distance, orientation)
        if mate1.reference_id != mate2.reference_id:
            inter_chromosomal += 1

        elif abs(mate2.pos - mate1.pos) < 20000:
            short_range += 1
        else:
            long_range += 1

        if orientation == 'inward':
            count_inward += 1
        elif orientation == 'outward':
            count_outward += 1
        elif orientation == 'same-strand-left':
            count_left += 1
        elif orientation == 'same-strand-right':
            count_right += 1

        for mate in [mate1, mate2]:
            # fill in coverage vector
            vec_start = max(0, mate.pos - mate_bin.begin) / binsize
            vec_end = min(len(coverage[mate_bin_id]), vec_start +
                          len(mate.seq) / binsize)
            coverage[mate_bin_id][vec_start:vec_end] += 1

        row.append(mate_bins[0])
        col.append(mate_bins[1])
        data.append(1)

        pair_added += 1

        # prepare data for bam output
        # set the flag to point that this data is paired
        mate1.flag |= 0x1
        mate2.flag |= 0x1

        # set one read as the first in pair and the
        # other as second
        mate1.flag |= 0x40
        mate2.flag |= 0x80

        # set chrom of mate
        mate1.mrnm = mate2.rname
        mate2.mrnm = mate1.rname

        # set position of mate
        mate1.mpos = mate2.pos
        mate2.mpos = mate1.pos

        out_bam.write(mate1)
        out_bam.write(mate2)

        if iter_num % 5e6 == 0:
            # every 5 million iterations append to the matrix
            # otherwise the row, col and data vectors continue growing and
            # for a large dataset the system could run out of memory
            if hic_matrix is None:
                hic_matrix = coo_matrix((data, (row, col)), shape=(matrix_size, matrix_size))
            else:
                hic_matrix += coo_matrix((data, (row, col)), shape=(matrix_size, matrix_size))
            row = []
            col = []
            data = []

    if hic_matrix is None:
        hic_matrix = coo_matrix((data, (row, col)), shape=(matrix_size, matrix_size))
    else:
        hic_matrix += coo_matrix((data, (row, col)), shape=(matrix_size, matrix_size))

    # the resulting matrix is only filled unevenly with some pairs
    # int the upper triangle and others in the lower triangle. To construct
    # the definite matrix I add the values from the upper and lower triangles
    # and subtract the diagonal to avoid double counting it.
    # The resulting matrix is symmetric.
    dia = dia_matrix(([hic_matrix.diagonal()], [0]), shape=hic_matrix.shape)
    hic_matrix = hic_matrix + hic_matrix.T - dia

    # extend bins such that they are next to each other
    bin_intervals = enlarge_bins(bin_intervals[:], chrom_sizes)

    # compute max bin coverage
    bin_max = []
    for cov in coverage:
        # bin_coverage.append(round(float(len(cov[cov > 0])) / len(cov), 3))
        if len(cov) == 0:
            bin_max.append(np.nan)
        else:
            bin_max.append(max(cov))

    chr_name_list, start_list, end_list = zip(*bin_intervals)
    bin_intervals = zip(chr_name_list, start_list, end_list, bin_max)
    hic_ma = hm.hiCMatrix()
    hic_ma.setMatrix(hic_matrix, cut_intervals=bin_intervals)

    args.outFileName.close()
    # removing the empty file. Otherwise the save method
    # will say that the file already exists.
    unlink(args.outFileName.name)

    hic_ma.save(args.outFileName.name)

    """
    if args.restrictionCutFile:
        # load the matrix to mask those
        # bins that most likely didn't
        # have a restriction site that was cutted

        # reload the matrix as a HiCMatrix object
        hic_matrix = hm.hiCMatrix(args.outFileName.name)

        hic_matrix.maskBins(get_poor_bins(bin_max))
        hic_matrix.save(args.outFileName.name)
    """
    if args.removeSelfLigation:
        msg = " (removed)"
    else:
        msg = " (not removed)"

    mappable_pairs = iter_num - one_mate_unmapped

    log_file_name = os.path.join(args.QCfolder, "QC.log")
    log_file = open(log_file_name, 'w')
    log_file.write("""
File\t{}\t\t
Pairs considered\t{}\t\t
Min rest. site distance\t{}\t\t
Max rest. site distance\t{}\t\t

""".format(args.outFileName.name, iter_num, args.minDistance,
           args.maxDistance))

    log_file.write("Pairs used\t{}\t({:.2f})\t({:.2f})\n".format(pair_added,
                                                                 100 * float(pair_added) / iter_num,
                                                                 100 * float(pair_added) / mappable_pairs))
    log_file.write("One mate unmapped\t{}\t({:.2f})\t({:.2f})\n".format(one_mate_unmapped,
                                                                        100 * float(one_mate_unmapped) / iter_num,
                                                                        100 * float(one_mate_unmapped) / mappable_pairs))

    log_file.write("One mate not unique\t{}\t({:.2f})\t({:.2f})\n".format(one_mate_not_unique,
                                                                          100 * float(one_mate_not_unique) / iter_num,
                                                                          100 * float(one_mate_not_unique) / mappable_pairs))

    log_file.write("One mate low quality\t{}\t({:.2f})\t({:.2f})\n".format(one_mate_low_quality,
                                                                           100 * float(one_mate_low_quality) / iter_num,
                                                                           100 * float(one_mate_low_quality) / mappable_pairs))

    log_file.write("dangling end\t{}\t({:.2f})\t({:.2f})\n".format(dangling_end,
                                                                   100 * float(dangling_end) / iter_num,
                                                                   100 * float(dangling_end) / mappable_pairs))

    log_file.write("self ligation{}\t{}\t({:.2f})\t({:.2f})\n".format(msg, self_ligation,
                                                                      100 * float(self_ligation) / iter_num,
                                                                      100 * float(self_ligation) / mappable_pairs))

    log_file.write("One mate not close to rest site\t{}\t({:.2f})\t({:.2f})\n".format(mate_not_close_to_rf,
                                                                                      100 * float(mate_not_close_to_rf) / iter_num,
                                                                                      100 * float(mate_not_close_to_rf) / mappable_pairs))

    log_file.write("same fragment (800 bp)\t{}\t({:.2f})\t({:.2f})\n".format(same_fragment,
                                                                             100 * float(same_fragment) / iter_num,
                                                                             100 * float(same_fragment) / mappable_pairs))
    log_file.write("self circle\t{}\t({:.2f})\t({:.2f})\n".format(self_circle,
                                                                  100 * float(self_circle) / iter_num,
                                                                  100 * float(self_circle) / mappable_pairs))
    log_file.write("duplicated pairs\t{}\t({:.2f})\t({:.2f})\n".format(duplicated_pairs,
                                                                       100 * float(duplicated_pairs) / iter_num,
                                                                       100 * float(duplicated_pairs) / mappable_pairs))
    if pair_added > 0:
        log_file.write("Of pairs used:\n")
        log_file.write("inter chromosomal\t{}\t({:.2f})\n".format(inter_chromosomal, 100 * float(inter_chromosomal) / pair_added))

        log_file.write("short range < 20kb\t{}\t({:.2f})\n".format(short_range, 100 * float(short_range) / pair_added))

        log_file.write("long range\t{}\t({:.2f})\n".format(long_range, 100 * float(long_range) / pair_added))

        log_file.write("inward pairs\t{}\t({:.2f})\n".format(count_inward, 100 * float(count_inward) / pair_added))

        log_file.write("outward pairs\t{}\t({:.2f})\n".format(count_outward, 100 * float(count_outward) / pair_added))

        log_file.write("left pairs\t{}\t({:.2f})\n".format(count_left, 100 * float(count_left) / pair_added))

        log_file.write("right pairs\t{}\t({:.2f})\n".format(count_right, 100 * float(count_right) / pair_added))

    log_file.close()
    QC.main("-l {} -o {}".format(log_file_name, args.QCfolder).split())


class Tester(object):
    def __init__(self):
        import os
        hic_test_data_dir = os.environ.get('HIC_TEST_DATA_DIR', False)
        if hic_test_data_dir:
            self.root = hic_test_data_dir
        else:
            self.root = os.path.dirname(os.path.abspath(__file__)) + "/test/test_data/"
        self.bam_file_1 = os.path.join(self.root, "hic.bam")



#!/usr/bin/env python
#-*- coding: utf-8 -*-

import argparse, sys
import numpy as np
import scipy.sparse
import scipy.stats
import multiprocessing

from hicexplorer import HiCMatrix
from hicexplorer.utilities import _fdr

import parserCommon

w_width = None
p_width = None

orig_matrix = None

pvalue = True

def parse_argumentss(args=None):
    """
    parse arguments
    """

    parent_parser = parserCommon.getParentArgParse()
    parser = argparse.ArgumentParser(
        parents=[parent_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Applies a transformation to a Hi-C matrix.')

    parser.add_argument('--originalMat',
                        help='File name containing the original hic matrix',
                        type=argparse.FileType('r'))

    parser.add_argument('--outFileName', '-o',
                        help='File name to save the resulting matrix ',
                        type=parserCommon.writableFile,
                        required=True)

    parser.add_argument(
                        '--chromosome',
                        help='Chromosome name to compute the values ',
                        required=True,
                        default=None)

    parser.add_argument('--numberOfProcessors',  '-p',
                        help='Number of processors to use ',
                        type=int,
                        default=1)

    parser.add_argument('--type',
                        help='output type. The options are "obs/exp" or "p-value".',
                        choices=['obs/exp', 'p-value'],
                        required=True)

    return parser.parse_args(args)


def compute_local_background_for_pixel(matrix, expected_matrix, row, col, correction_factors):
    """
    Given a corrected Hi-C matrix and a 'expected values' matrix this function returns
    either the obs/exp ratio or the p-value after considering the local background using
    the shapes proposed by Rao et al (lower_left, donut, horizontal and vertical). It chooses
    the shape that results in the lower ratio (or least significant p-value)
    :param matrix: scipy sparse csr corrected Hi-C matrix
    :param expected_matrix: scipy sparse csr expected values matrix
    :param row: row position of pixel to evaluate
    :param col: col position of pixel to evaluate
    :param correction_factors: Correction factors used to correct the Hi-C matrix

    :return:
    """

    global p_width, w_width
    lambda_ = [expected_matrix[row, col]]
    for what in ['lower_left', 'donut', 'horizontal', 'vertical']:
        counts = get_matrix_section_mean(matrix, row, col, w_width, p_width, what)

        exp = get_matrix_section_mean(expected_matrix, row, col, w_width, p_width, what)

        if np.isnan(counts):
            # close to the matrix borders and close to the diagonal
            # the value of the counts is nan.
            lambda_.append(np.nan)
            return np.nan
        if counts == 0 and exp > 0:
            counts = exp

        lambda_.append((float(counts)/exp) * expected_matrix[row, col])

    # get the highest of the computed lambdas
    lambda_ = max(lambda_)

    if ARGS.type == 'p-value':

        pixel_value = np.rint(matrix[row, col] * correction_factors[row] * correction_factors[col])
        value = -1 * scipy.stats.poisson.logsf(pixel_value,
                                               lambda_ * correction_factors[row] * correction_factors[col])
    else:
        value = float(matrix[row, col]) / lambda_

    return value
        
    
def get_matrix_section_mean(matrix, row, col, w_width, p_width, what='lower_left'):
    """
    computes sub matrices based on Rao et al. 2014
    :param matrix: Hi-C matrix
    :param i: row idx
    :param j: col idx
    :param what: string, type of matrix. Options are
            lower_left, donut, horizontal, vertical
    :return: mean value of counts found in the given
    matrix section

    >>> w_width = 5
    >>> p_width =2
    >>> matrix = np.zeros((13,13))
    >>> matrix[1:12, 1:12] = 1
    >>> matrix[4:9,4:9] = 5
    >>> matrix[6,6] = 10
    >>> #print matrix
    >>> get_matrix_section_mean(scipy.sparse.csr_matrix(matrix), 6, 6, w_width, p_width, 'donut')
    1.0
    >>> matrix = np.zeros((13,13))
    >>> matrix[1:12, 1:12] = 1
    >>> matrix[7:12,1:6] = 3
    >>> matrix[4:9,4:9] = 5
    >>> matrix[6,6] = 10
    >>> # print matrix
    >>> get_matrix_section_mean(scipy.sparse.csr_matrix(matrix), 6, 6, w_width, p_width, 'lower_left')
    3.0
    >>> get_matrix_section_mean(scipy.sparse.csr_matrix(matrix), 6, 6, w_width, p_width, 'horizontal')
    1.3333333333333333
    >>> get_matrix_section_mean(scipy.sparse.csr_matrix(matrix), 6, 6, w_width, p_width, 'vertical')
    1.3333333333333333

    """


    M = matrix.shape[0]
    # check that the area to check is not too close to the borders or
    # to the diagonal
    if row-w_width < 0 or col-w_width < 0 or \
        row+w_width + 1 > M or col+w_width + 1 > M or \
        row+w_width+1 > col - w_width:

        return np.nan

    if what == 'lower_left':
        values = np.concatenate([matrix[row+1:row+w_width+1,
                                  col-w_width:col-p_width].data,
                                 matrix[row+p_width+1:row+w_width+1,
                                  col-p_width:col-1].data])

    elif what == 'donut':
        # donut values are computed by first selecting a sub-matrix and
        # then subtracting the regions that are not interesting?

        donut_matrix = matrix[row-w_width:row+w_width+1,
                              col-w_width:col+w_width+1].todense()

        # remove the donut center:
        donut_matrix[w_width-p_width:w_width+p_width+1,
                     w_width-p_width:w_width+p_width+1] = 0
        # remove horizontal and vertical parts
        donut_matrix[w_width,:] = 0
        donut_matrix[:, w_width] = 0

        values = donut_matrix[np.nonzero(donut_matrix)]

    elif what == 'vertical':
        values = np.concatenate([matrix[row-w_width:row-p_width,
                                  col-1:col+1+1].data,
                                 matrix[row+p_width+1:row+w_width+1,
                                  col-1:col+1+1].data])

    elif what == 'horizontal':
        values = np.concatenate([matrix[row-1:row+1+1,
                                  col-w_width:col-p_width].data,
                                 matrix[row-1:row+1+1,
                                  col+p_width+1:col+w_width+1].data])

    return values.sum()

def get_ratio_wrapper(args):
    return get_ratio_worker(*args)
    
def get_ratio_worker(triu_ma, csr_matrix, expected_ma, bin_size,
                     correction_factors, idx_array):

    print "Processing from {} to {}".format(min(idx_array), max(idx_array))
    #    for idx, pixel_value in enumerate(triu_ma.-wdata):
    transf_ma = np.zeros(len(idx_array))

    for idx in idx_array:
        # compute lower_left value
        row = triu_ma.row[idx]
        col = triu_ma.col[idx]

        transf_ma[idx-min(idx_array)] = compute_local_background_for_pixel(csr_matrix,
                                                                           expected_ma, row, col, correction_factors)

        """
        lower_left_mean = get_matrix_section_mean(csr_matrix, expected_ma, row, col,
                                                  bin_size, correction_factors, what='lower_left')
        donut_mean  = get_matrix_section_mean(csr_matrix, expected_ma, row, col,
                                              bin_size,  correction_factors, what='donut')
        horizontal_mean  = get_matrix_section_mean(csr_matrix, expected_ma, row, col,
                                                   bin_size, correction_factors, what='horizontal')
        vertical_mean  = get_matrix_section_mean(csr_matrix, expected_ma, row, col,
                                                 bin_size,  correction_factors, what='vertical')
                                                 


        transf_ma[idx-min(idx_array)] = float(pixel_value) / max(lower_left_mean, donut_mean,
                                                   horizontal_mean, vertical_mean)
        """
        if idx % 100 == 0:
            sys.stderr.write(".") #  print a . to show progress

    return transf_ma

def main(args):

    hicma = HiCMatrix.hiCMatrix(args.matrix)

    hicma.keepOnlyTheseChr(args.chromosome)

    if hicma.correction_factors is None:
        exit("matrix {} does not have correction factors. Iterative corrected matrix required.".format(args.matrix))

    # check that the matrix has bins of same size
    # otherwise try to adjust the bins to
    # to match a regular binning
    chrom, start, end, extra = zip(*hicma.cut_intervals)

    median = int(np.median(np.diff(start)))
    diff = np.array(end) - np.array(start)
    # check if the bin size is homogeneous
    if len(np.flatnonzero(diff != median)) > (len(diff) * 0.01):
        # set the start position of a bin to the closest multiple
        # of the median
        def snap_nearest_multiple(start_x, m):
            resi = [-1*(start_x % m), -start_x % m]
            return start_x + resi[np.argmin(np.abs(resi))]
        start = [snap_nearest_multiple(x, median) for x in start]
        end = [snap_nearest_multiple(x, median) for x in end]
        cut_intervals = zip(chrom, start, end, extra)
        sys.stderr.write('[getCountsByDistance] Bin size is not '
                         'homogeneous, setting \n'
                         'the bin distance to the median: {}\n'.format(median))

    # TODO make generic for all chromosomes
    # get counts by distance for chrX
    counts_by_distance = hicma.getCountsByDistance(per_chr=True)[args.chromosome]

    # construct expected matrix
    diagonals_array = []
    M = hicma.matrix.shape[0]
    for k in np.sort(counts_by_distance.keys()):
        diagonals_array.append(np.repeat(counts_by_distance[k].mean(), M))

    diagonals_array = np.vstack(diagonals_array)
    # the expected value matrix is not sparse but
    # the constructor is useful to add the diagonals
    # each containing the diagonal mean in each position.
    # In other words, the following line is a hack to
    # to convert rows in a matrix to diagonals.
    exp_mat = scipy.sparse.dia_matrix((diagonals_array, np.arange(0, len(diagonals_array))), shape=(M,M))

    # make the matrix symmetric
    exp_mat = exp_mat + scipy.sparse.triu(exp_mat, 1).T
    exp_mat = exp_mat.tocsr()
    import copy
    exp_mat_obj = copy.copy(hicma)
    exp_mat_obj.setMatrixValues(exp_mat)
    exp_mat_obj.maskBins(hicma.nan_bins)

    hicma.maskBins(hicma.nan_bins)

    # use only upper part of the matrix
    # and convert to coo
    triu_ma = scipy.sparse.triu(hicma.matrix, format='coo')

    global bin_size, w_width, p_width

    bin_size = hicma.getBinSize()
    #kb_20_in_bins = max(1, 20000/bin_size)
    #w_width = int(kb_20_in_bins * 8)
    #p_width =  max(1, int(kb_20_in_bins * 4))
    # for 25 kb
    w_width = 3
    p_width =  1

    #w_width = 5
    #p_width =  2

    #import ipdb;ipdb.set_trace()
    #compute_local_background_for_pixel(hicma.matrix, exp_mat, 463, 474, bin_size, hicma.correction_factors, what='lower_left')

    # matrix to hold the results (initially just an array)
    transf_ma = np.zeros(len(triu_ma.data))
#    import ipdb;ipb.set_trace()
    # this loop is to process o>nly pixels with data
    num_processors = args.numberOfProcessors
    pool = multiprocessing.Pool(num_processors)
    func = get_ratio_wrapper
    TASKS = []
    for idx_array in np.array_split(np.arange(len(triu_ma.data)), num_processors):
        TASKS.append((triu_ma, hicma.matrix, exp_mat_obj.matrix, bin_size, 
                      hicma.correction_factors, idx_array))

    if num_processors > 1:
        res = pool.map_async(func, TASKS).get(9999999)
    else:
        res = map(func, TASKS)
#    import ipdb;ipdb.set_trace()
    transf_ma = np.concatenate(res)
    # set the new values back into the original matrix
    if args.type == 'p-value':
        sys.stderr.write("computing FDR\n")
        #triu_ma.data = -np.log(_fdr(np.exp(-transf_ma)))
        transf_ma = np.exp(-transf_ma) * len(transf_ma)
        transf_ma[transf_ma > 1] = 1
        triu_ma.data = -np.log(transf_ma)

    else:
            triu_ma.data = transf_ma
    # fill the lower triangle
    triu_ma = triu_ma + scipy.sparse.triu(triu_ma, 1).T
    triu_ma = triu_ma.tocsr()
    triu_ma.eliminate_zeros()

    hicma.setMatrixValues(triu_ma)
    hicma.restoreMaskedBins()

    hicma.save(args.outFileName)

if __name__ == "__main__":
    ARGS = parse_argumentss()
    main(ARGS)




"""

[hic]
file = /data/manke/group/ramirez/HiC-ThomasLing/data/hic_data/s2/hic-norm/RF/merge.npz
title =
colormap = RdYlBu_r
depth = 100000
#min_value =2.8
#max_value = 3.0
transform = log1p
boundaries_file = conductance_vs_hic/boundaries_all.bed
x labels = yes
type = arcplot
type = interaction
#optional in case it can not be guessed by the file ending
file_type = hic_matrix
# show masked bins plots as white lines
# those bins that were not used during the correction
# the default is to extend neighboring bins to
# obtain an aesthetically pleasant output
show_masked_bins = yes

[x-axis]
#optional
fontsize=20
# default is bottom meaning below the axis line
where=top

[spacer]

[bigwig]
file = /data/manke/group/ramirez/HiC-ThomasLing/data/external/Graveley_mRNA-seq/GSM390060_Kc167-4_spa.bw
title = Kc RNA-seq (Cherbas et al.)
color = black
min_value = 0
#max_value = auto
width = 1.5
number of bins = 500
nans to zeros = True
#optional in case it can not be guessed by the file ending
file_type = bigwig

[simple bed]
file = file.bed
title = peaks
color = read
width = 0.5
# optional. If not given is guessed from the file ending
file_type = bed

[bed genes]
# example of a genes track
# has the same options as a simple
# bed but if the type=genes is given
# the the file is interpreted as gene
# file. If the bed file contains the exon
# structure then this is plotted. Otherwise
# a region **with direction** is plotted.
file = genes.bed
title = genes
color = darkblue
#if color is a valid colormap, then the score is mapped
# to the colormap
#color = RdBlGn
width = 5
# to turn off/on printing of labels
labels = off
# optional. If not given is guessed from the file ending
file_type = bed
# optional: font size can be given if default are not good
fontsize = 10
# optional
# display = collapsed
# display = interleaved
#optional, default is black
#border_color = black
# style to plot the genes when they have exon information
#style = UCSC
#style = flybase
# maximum number of gene rows to be plotted. This
# field is useful to limit large number of close genes
# to be printed over many rows. When several images want
# to be combined this must be set to get equal size
# genes in all images
#gene rows = 10
# by default the ymax is the number of
# rows occupied by the genes in the region plotted. However,
# by setting this option, the global maximum is used instead.
# This is useful to combine images that are all consistent and
# have the same number of rows.
#global max row = yes


[chrom states]
# this is a case of a bed file that is plotted 'collapsed'
# useful to plot chromatin states if the bed file contains
# the color to plot
file = chromatinStates.bed
title = chromatin states
# color is replaced by the color in the bed file
# in this case
color = black
# default behaviour when plotting intervals from a
# bed file is to 'expand' them such that they
# do not overlap. The display = collapsed
# directive overlaps the intervals.
display = collapsed
width = 0.3

[bedgraph]
file = file.bg
title = bedgraph track
color = green
width = 0.2
# optional, otherwise guseed from file ending
file_type = bedgraph


[bedgraph matrix]
# a bedgraph matrix file is like a bedgraph, except that per bin there
# are more than one value separated by tab: E.g.
# chrX	18279	40131	0.399113	0.364118	0.320857	0.274307
# chrX	40132	54262	0.479340	0.425471	0.366541	0.324736
file = spectra_conductance.bm
title = conductance spectra
width = 1.5
orientation = inverted
min_value = 0.10
max_value = 0.70
# if type is set as lines, then the TAD score lines are drawn instead
# of the matrix
# set to lines if a heatmap representing the matrix
# is not wanted
type = lines
file_type = bedgraph_matrix
plot horizontal lines=False

[dendrogram]
# a dendrogram is saved in a bed file that contains
# further fields defining the linkage.
file =  linkage.bed
title = dendrogram
width = 2
orientation = inverted
# required on this case. Otherwise
# the bed file will be plotted as regions
file_type = dendrogram
# y positions to draw horizontal lines
hlines = 0.2 0.3

[vlines]
# vertical dotted lines from the top to the bottom of the figure
# can be drawn. For this a bed file is required
# but only the first two columns (chromosome name and start
# are used.
# vlines can also be given at the command line as a list
# of genomic positions. However, sometimes to give a file
# is more convenient in case many lines want to be plotted.
file = regions.bed
type = vlines


"""

from __future__ import division
import sys
import argparse
import matplotlib
matplotlib.use('Agg')

import hicexplorer.trackPlot
from hicexplorer._version import __version__

DEFAULT_BED_COLOR = '#1f78b4'
DEFAULT_BIGWIG_COLOR = '#33a02c'
DEFAULT_BEDGRAPH_COLOR = '#a6cee3'
DEFAULT_MATRIX_COLORMAP = 'RdYlBu_r'
DEFAULT_TRACK_HEIGHT = 3  # in centimeters
DEFAULT_FIGURE_WIDTH = 40  # in centimeters
# proportion of width dedicated to (figure, legends)
# DEFAULT_WIDTH_RATIOS = (0.95, 0.05)
DEFAULT_MARGINS = {'left': 0.04, 'right': 0.92, 'bottom': 0.12, 'top': 0.9}


def parse_arguments(args=None):
    parser = argparse.ArgumentParser(
        description='Plots the diagonal,  and some values close to '
        'the diagonal of a  HiC matrix. The diagonal of the matrix is '
        'plotted horizontally for a region. I will not draw the diagonal '
        'for the whole chromosome',
        usage="%(prog)s --tracks tracks.ini --region chr1:1000000-4000000 -o image.png")

    parser.add_argument('--tracks',
                        help='File containing the instructions to plot the tracks ',
                        type=argparse.FileType('r'),
                        required=True,
                        )

    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument('--region',
                       help='Region to plot, the format is chr:start-end')

    group.add_argument('--BED',
                       help='Instead of a region, a file containing the regions to plot, in BED format, '
                       'can be given. If this is the case, multiple files will be created using a prefix '
                       'the value of --outFileName',
                       type=argparse.FileType('r')
                       )

    parser.add_argument('--width',
                        help='figure width in centimeters',
                        type=float,
                        default=DEFAULT_FIGURE_WIDTH)

    parser.add_argument('--height',
                        help='Figure height in centimeters. If not given, the figure height is computed '
                             'based on the widths of the tracks and on the depth '
                             'of the Hi-C tracks, if any. Setting this '
                             'this parameter can cause the Hi-C tracks to look collapsed or expanded.',
                        type=float)

    parser.add_argument('--title', '-t',
                        help='Plot title',
                        required=False)

    parser.add_argument('--scoreName', '-s',
                        help='Score name',
                        required=False)

    parser.add_argument('--outFileName', '-out',
                        help='File name to save the image, file prefix in case multiple images '
                             'are stored',
                        required=True)

    parser.add_argument('--zMax',
                        help='zMax',
                        type=float,
                        default=None)

    parser.add_argument('--vlines',
                        help='Genomic cooordindates separated by space. E.g. '
                        '--vlines 150000 3000000 124838433 ',
                        type=int,
                        nargs='+'
                        )

    parser.add_argument('--fontSize',
                        help='Font size for the labels of the plot',
                        type=float,
                        )

    parser.add_argument('--dpi',
                        help='Resolution for the image in case the'
                             'ouput is a raster graphics image (e.g png, jpg)',
                        type=int,
                        default=72
                        )

    parser.add_argument('--trackLabelFraction',
                        help='By default the space dedicated to the track labels is 0.05 of the'
                             'plot width. This fraction can be changed with this parameter if needed.',
                        default=0.05,
                        type=float)

    parser.add_argument('--version', action='version',
                        version='%(prog)s {}'.format(__version__))

    return parser


def get_region(region_string):
    """
    splits a region string into
    a chrom, start_region, end_region tuple
    The region_string format is chr:start-end
    """
    if region_string:
        region_string = region_string.translate(
            None, ",.;|!{}()").replace("-", ":")
        region = region_string.split(":")
        chrom = region[0]
        try:
            region_start = int(region[1])
        except IndexError:
            region_start = 0
        try:
            region_end = int(region[2])
        except IndexError:
            region_end = 1e15  # a huge number

        if region_end <= region_start:
            exit("Please check that the region end is larger than the region start.\n"
                 "Values given:\nstart: {}\nend: {}\n".format(region_start, region_end))

        return chrom, region_start, region_end


def main(args=None):

    args = parse_arguments().parse_args(args)
    trp = hicexplorer.trackPlot.PlotTracks(args.tracks.name, args.width, fig_height=args.height,
                                           fontsize=args.fontSize, dpi=args.dpi,
                                           track_label_width=args.trackLabelFraction)

    if args.BED:
        count = 0
        for line in args.BED.readlines():
            count += 1
            try:
                chrom, start, end = line.strip().split('\t')[0:3]
            except ValueError:
                continue
            try:
                start, end = map(int, [start, end])
            except ValueError as detail:
                sys.stderr.write("Invalid value found at line\t{}\t. {}\n".format(line, detail))
            file_name = "{}_{}:{}-{}".format(args.outFileName, chrom, start, end)
            if end - start < 200000:
                start -= 100000
                end += 100000
            sys.stderr.write("saving {}'\n".format(file_name))
            trp.plot(file_name, chrom, start, end, title=args.title)
    else:
        region = get_region(args.region)
        trp.plot(args.outFileName, *region, title=args.title)



import argparse
import sys
import os
import numpy as np
from scipy.sparse import triu
from scipy.stats import pearsonr, spearmanr

import hicexplorer.HiCMatrix as hm
from hicexplorer._version import __version__

# for plotting
from matplotlib import use as mplt_use
mplt_use('Agg')

import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.ticker import FixedLocator


def parse_arguments(args=None):

    heatmap_parser = heatmap_options()

    parser = argparse.ArgumentParser(
        parents=[heatmap_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Computes pairwise correlations between hic matrices data. '
        'The correlation is computed taking the values from each pair '
        'of matrices and discarding values that are zero in both matrices.')

    # define the arguments
    parser.add_argument('--matrices', '-m',
                        help='matrices to correlate.',
                        nargs='+',
                        required=True)

    # define the arguments
    parser.add_argument('--method',
                        help='Correlation method to use.',
                        choices=['pearson', 'spearman'],
                        default='pearson')

    parser.add_argument('--log1p',
                        help='Use the log1p of the matrix values.',
                        action='store_true')

    parser.add_argument('--labels', '-l',
                        metavar='sample1 sample2',
                        help='User defined labels instead of default labels '
                        'from file names. '
                        'Multiple labels have to be separated by space, e.g. '
                        '--labels sample1 sample2 sample3',
                        nargs='+')

    parser.add_argument('--range',
                        help='In bp with the format low_range:high_range, '
                        'for example 1000000:2000000. If --range is given '
                        'only counts within the range are considered.')

    parser.add_argument('--outFileNameHeatmap', '-oh',
                        help='File name to save the resulting heatmap plot',
                        required=True)

    parser.add_argument('--outFileNameScatter', '-os',
                        help='File name to save the resulting scatter plot',
                        required=True)

    parser.add_argument('--chromosomes',
                        help='List of chromosomes to be included in the '
                        'correlation.',
                        default=None,
                        nargs='+')
    parser.add_argument('--version', action='version',
                        version='%(prog)s {}'.format(__version__))

    return parser


def heatmap_options():
    """
    Options for generating the correlation heat map
    """
    parser = argparse.ArgumentParser(add_help=False)
    heatmap = parser.add_argument_group('Heat map options')

    heatmap.add_argument('--zMin', '-min',
                         default=None,
                         help='Minimum value for the heat map intensities. '
                              'If not specified the value is set automatically')
    heatmap.add_argument('--zMax', '-max',
                         default=None,
                         help='Maximum value for the heat map intensities.'
                              'If not specified the value is set automatically')

    heatmap.add_argument(
        '--colorMap', default='jet',
        metavar='',
        help='Color map to use for the heatmap. Available values can be '
             'seen here: '
             'http://matplotlib.org/examples/color/colormaps_reference.html')

    heatmap.add_argument('--plotFileFormat',
                         metavar='FILETYPE',
                         help='Image format type. If given, this option '
                              'overrides the image format based on the plotFile '
                              'ending. The available options are: png, emf, '
                              'eps, pdf and svg.',
                         choices=['png', 'pdf', 'svg', 'eps', 'emf'])

    heatmap.add_argument('--plotNumbers',
                         help='If set, then the correlation number is plotted '
                              'on top of the heatmap.',
                         action='store_true',
                         required=False)
    return parser


def plot_correlation(corr_matrix, labels, plot_filename, vmax=None,
                     vmin=None, colormap='Reds', image_format=None):
    import scipy.cluster.hierarchy as sch
    num_rows = corr_matrix.shape[0]

    # set the minimum and maximum values
    if vmax is None:
        vmax = 1
    if vmin is None:
        vmin = 0 if corr_matrix.min() >= 0 else -1

    # Compute and plot dendrogram.
    fig = plt.figure(figsize=(10.5, 9.5))
    axdendro = fig.add_axes([0.02, 0.1, 0.1, 0.7])
    axdendro.set_axis_off()
    y_var = sch.linkage(corr_matrix, method='complete')
    z_var = sch.dendrogram(y_var, orientation='right',
                           link_color_func=lambda k: 'black')
    axdendro.set_xticks([])
    axdendro.set_yticks([])
    cmap = plt.get_cmap(colormap)

    # this line simply makes a new cmap, based on the original
    # colormap that goes from 0.0 to 0.9
    # This is done to avoid colors that
    # are too dark at the end of the range that do not offer
    # a good contrast between the correlation numbers that are
    # plotted on black.
    cmap = cmap.from_list(colormap + "clipped", cmap([0.0, 0.8]))
    # Plot distance matrix.
    axmatrix = fig.add_axes([0.13, 0.1, 0.6, 0.7])
    index = z_var['leaves']
    corr_matrix = corr_matrix[index, :]
    corr_matrix = corr_matrix[:, index]
    img_mat = axmatrix.matshow(corr_matrix, aspect='equal', origin='lower',
                               cmap=cmap, extent=(0, num_rows, 0, num_rows),
                               vmax=vmax, vmin=vmin)
    axmatrix.yaxis.tick_right()
    axmatrix.set_yticks(np.arange(corr_matrix.shape[0]) + 0.5)
    axmatrix.set_yticklabels(np.array(labels).astype('str')[index],
                             fontsize=14)

    axmatrix.set_xticks(np.arange(corr_matrix.shape[0]) + 0.5)
    axmatrix.set_xticklabels(np.array(labels).astype('str')[index],
                             fontsize=14,
                             rotation=45,
                             ha='left')

#    axmatrix.set_xticks([])
    # Plot colorbar.
    axcolor = fig.add_axes([0.13, 0.065, 0.6, 0.02])
    plt.colorbar(img_mat, cax=axcolor, orientation='horizontal')
    for row in range(num_rows):
        for col in range(num_rows):
            axmatrix.text(row + 0.5, col + 0.5,
                          "{:.2f}".format(corr_matrix[row, col]),
                          ha='center', va='center')

    fig.savefig(plot_filename, format=image_format)


def get_vectors(mat1, mat2):
    """
    Uses sparse matrix tricks to convert
    into a vector the matrix values such
    that zero values that appear in only
    one of the matrices is kept. But
    zeros in two matrices are removed

    Requires two sparse matrices as input
    """
    assert mat1.shape == mat2.shape, "Matrices have different shapes. "\
        "Computation of correlation is not possible."

    # create a new matrix that is the sum of the two
    # matrices to compare. The goal is to have
    # a matrix that contains all the positions
    # that are non-zero in both matrices
    _mat = mat1 + mat2

    # add one to each element in the new matrix
    _mat.data += 1

    # get a vector of the values in mat1 from
    # _mat
    values1 = (_mat - mat1).data - 1

    # get a vector of the values in mat2 from
    # _mat
    values2 = (_mat - mat2).data - 1

    return values1, values2


def main():

    args = parse_arguments().parse_args()

    if args.labels and len(args.matrices) != len(args.labels):
        print "The number of labels does not match the number of matrices."
        exit(0)
    if not args.labels:
        args.labels = map(lambda x: os.path.basename(x), args.matrices)

    num_files = len(args.matrices)
    map(lambda x: os.path.basename(x), args.matrices)
    # initialize results matrix
    results = np.zeros((num_files, num_files), dtype='float')

    rows, cols = np.triu_indices(num_files)
    correlation_opts = {'spearman': spearmanr,
                        'pearson': pearsonr}
    hic_mat_list = []
    max_value = None
    min_value = None
    all_mat = None
    all_nan = []

    for matrix in args.matrices:
        sys.stderr.write("loading hic matrix {}\n".format(matrix))
        _mat = hm.hiCMatrix(matrix)
        _mat.diagflat(0)
        sys.stderr.write("restore masked bins {}\n".format(matrix))
        if args.chromosomes:
            _mat.keepOnlyTheseChr(args.chromosomes)
        _mat.filterOutInterChrCounts()
        bin_size = _mat.getBinSize()
        all_nan = np.unique(np.concatenate([all_nan, _mat.nan_bins]))

        _mat = triu(_mat.matrix, k=0, format='csr')
        """
        # estimate noise level
        noise_cutoff = np.percentile(_mat.data,70)
        print "noise threshold set to: {}".format(noise_cutoff)
        _mat.data[_mat.data<=noise_cutoff] = 0
        _mat.eliminate_zeros()
        """
        if args.range:
            min_dist, max_dist = args.range.split(":")
            min_dist = int(min_dist) / bin_size
            max_dist = int(max_dist) / bin_size
            _mat_coo = _mat.tocoo()
            dist = _mat_coo.col - _mat_coo.row
            keep = np.flatnonzero((dist <= max_dist) & (dist >= min_dist))
            _mat_coo.data = _mat_coo.data[keep]
            _mat_coo.row = _mat_coo.row[keep]
            _mat_coo.col = _mat_coo.col[keep]
            _mat = _mat_coo.tocsr()
        if args.log1p:
            _mat.data = np.log1p(_mat.data)
        if all_mat is None:
            all_mat = _mat
        else:
            all_mat = all_mat + _mat

        if max_value is None or max_value < _mat.data.max():
            max_value = _mat.data.max()
        if min_value is None or min_value > _mat.data.min():
            min_value = _mat.data.min()

        hic_mat_list.append(_mat)

    # remove nan bins
    rows_keep = cols_keep = np.delete(range(all_mat.shape[1]), all_nan)
    all_mat = all_mat[rows_keep, :][:, cols_keep]

    # make large matrix to correlate by
    # using sparse matrix tricks

    print len(all_nan)
    big_mat = None
    for mat in hic_mat_list:
        mat = mat[rows_keep, :][:, cols_keep]
        sample_vector = (mat + all_mat).data - all_mat.data
        if big_mat is None:
            big_mat = sample_vector
        else:
            big_mat = np.vstack([big_mat, sample_vector])

    # take the transpose such that columns represent each of the samples
    big_mat = np.ma.masked_invalid(big_mat).T

    grids = gridspec.GridSpec(num_files, num_files)
    grids.update(wspace=0, hspace=0)
    fig = plt.figure(figsize=(2 * num_files, 2 * num_files))
    plt.rcParams['font.size'] = 8.0

    min_value = int(big_mat.min())
    max_value = int(big_mat.max())
    if (min_value % 2 == 0 and max_value % 2 == 0) or \
            (min_value % 1 == 0 and max_value % 2 == 1):
        # make one value odd and the other even
        max_value += 1

    if args.log1p:
        major_locator = FixedLocator(range(min_value, max_value, 2))
        minor_locator = FixedLocator(range(min_value, max_value, 1))

    for index in xrange(len(rows)):
        row = rows[index]
        col = cols[index]
        if row == col:
            results[row, col] = 1

            # add titles as
            # empty plot in the diagonal
            ax = fig.add_subplot(grids[row, col])
            ax.text(0.6, 0.6, args.labels[row],
                    verticalalignment='center',
                    horizontalalignment='center',
                    fontsize=10, fontweight='bold',
                    transform=ax.transAxes)
            ax.set_axis_off()
            continue

        sys.stderr.write("comparing {} and {}\n".format(args.matrices[row],
                                                        args.matrices[col]))

        # remove cases in which both are zero or one is zero and
        # the other is one
        _mat = big_mat[:, [row, col]]
        _mat = _mat[_mat.sum(axis=1) > 1, :]
        vector1 = _mat[:, 0]
        vector2 = _mat[:, 1]

        results[row, col] = correlation_opts[args.method](vector1, vector2)[0]

        # scatter plots
        ax = fig.add_subplot(grids[row, col])
        if args.log1p:
            ax.xaxis.set_major_locator(major_locator)
            ax.xaxis.set_minor_locator(minor_locator)
            ax.yaxis.set_major_locator(major_locator)
            ax.yaxis.set_minor_locator(minor_locator)

        ax.text(0.2, 0.8, "{}={:.2f}".format(args.method,
                                             results[row, col]),
                horizontalalignment='left',
                transform=ax.transAxes)
        ax.get_yaxis().set_tick_params(
            which='both',
            left='off',
            right='off',
            direction='out')

        ax.get_xaxis().set_tick_params(
            which='both',
            top='off',
            bottom='off',
            direction='out')

        if col != num_files - 1:
            ax.set_yticklabels([])
        else:
            ax.yaxis.tick_right()
            ax.get_yaxis().set_tick_params(
                which='both',
                left='off',
                right='on',
                direction='out')
        if col - row == 1:
            ax.xaxis.tick_bottom()
            ax.get_xaxis().set_tick_params(
                which='both',
                top='off',
                bottom='on',
                direction='out')
        else:
            ax.set_xticklabels([])

        ax.hist2d(vector1, vector2, bins=200, cmin=0.1)
        # downsample for plotting
#        choice_idx = np.random.randint(0,len(vector1),min(len(vector1), 500000))
#        ax.plot(vector1[choice_idx], vector2[choice_idx], '.', markersize=1,
#                 alpha=0.2, color='darkblue',
#                 markeredgecolor='none')


#        ax.set_ylim(min_value, max_value)
#        ax.set_xlim(min_value,max_value)
#        ax.set_ylim(-1, ax.get_ylim()[1])
#        ax.set_xlim(-1, ax.get_xlim()[1])
#        ax.set_ylim(min_value, ax.get_ylim()[1])
#        ax.set_xlim(min_value, ax.get_xlim()[1])
    fig.tight_layout()
    print "saving {}".format(args.outFileNameScatter)
    fig.savefig(args.outFileNameScatter, bbox_inches='tight')

    results = results + np.triu(results, 1).T
    plot_correlation(results, args.labels,
                     args.outFileNameHeatmap,
                     args.zMax,
                     args.zMin,
                     args.colorMap,
                     image_format=args.plotFileFormat)
#                    plot_numbers=args.plotNumbers)



import numpy as np
import pandas as pd
from scipy.sparse import coo_matrix

def getLibermanBins(filenameList, chrnameList, pandas = True):
        """
        Reads a list of txt file in liberman's format and returns
        cut intervals and matrix. Each file is seperated by chr name
        and contains: locus1,locus2,and contact score seperated by tab.
        """

        data = np.zeros([0,4])
        dim = 0
        ## Create empty row, col, value for the matrix

        row = np.array([]).astype("int")
        col = np.array([]).astype("int")
        value = np.array([])
        cut_intervals = []
        resolution = None

        ## for each chr, append the row, col, value to the first one. Extend the dim
        for i in range(0, len(filenameList)):
            if pandas == True:
                chrd = pd.read_csv(filenameList[i], sep = "\t", header=None)
                chrdata = chrd.as_matrix()
            else:
                print "Pandas unavailable. Reading files using numpy (slower).."
                chrdata = np.loadtxt(filenameList[i])

            # define resolution as the median of the difference of the rows
            # in the data table.
            if resolution is None:
                resolution = np.median(np.diff(np.unique(np.sort(chrdata[:,1])))).astype(int)

            chrcol = (chrdata[:, 1] / resolution).astype(int)
            chrrow = (chrdata[:, 0] / resolution).astype(int)

            chrdim = max(max(chrcol), max(chrrow)) + 1
            row = np.concatenate([row, chrrow + dim])
            col = np.concatenate([col, chrcol + dim])
            value = np.concatenate([value, chrdata[:, 2]])
            dim = dim + chrdim
            import ipdb;ipdb.set_trace()
            for _bin in range(chrdim):
                cut_intervals.append((chrnameList[i], _bin * resolution, (_bin + 1) * resolution))

        final_mat = coo_matrix((value, (row, col)), shape=(dim, dim))

        liberman_data = dict(cut_intervals=cut_intervals, matrix=final_mat)
        return liberman_data



matrix = getLibermanBins(['/data/akhtar/bhardwaj/2016_HiC/test/chr1', '/data/akhtar/bhardwaj/2016_HiC/test/chr2'],
                         ['1', '2'])

import ipdb;ipdb.set_trace()


import argparse
from hicexplorer import HiCMatrix as hm
from hicexplorer._version import __version__


def parse_arguments(args=None):

    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     description=('Prints information about a matrix including size, '
                                                  'number of elements, sum of elements, etc.'))

    parser.add_argument('--matrices', '-m',
                        help='matrices to add. Must have the same shape.',
                        metavar='.h5 of .npz file format',
                        nargs='+',
                        required=True)

    parser.add_argument('--version', action='version',
                        version='%(prog)s {}'.format(__version__))

    return parser


def main():

    args = parse_arguments().parse_args()
    for matrix in args.matrices:
        print("File:\t{}".format(matrix))

        hic_ma = hm.hiCMatrix(matrix)
        size = hic_ma.matrix.shape[0]
        num_non_zero = hic_ma.matrix.nnz
        sum_elements = hic_ma.matrix.sum() / 2
        bin_length = hic_ma.getBinSize()
        num_nan_bins = len(hic_ma.nan_bins)
        min_non_zero = hic_ma.matrix.data.min()
        max_non_zero = hic_ma.matrix.data.max()
        chromosomes = hic_ma.chrBinBoundaries.keys()

        print("Size:\t{:,}".format(size))
        print("Sum:\t{:,}".format(sum_elements))
        print("Bin_length:\t{}".format(bin_length))
        print("Chromosomes:\t{}".format(", ".join(chromosomes)))
        print("Non-zero elements:\t{:,}".format(num_non_zero))
        print("Minimum (non zero):\t{}".format(min_non_zero))
        print("Maximum:\t{}".format(max_non_zero))
        print("NaN bins:\t{}".format(num_nan_bins))
        print("")



from __future__ import division
import argparse
import numpy as np
import sys
from hicexplorer import HiCMatrix
import hicexplorer.parserCommon
from hicexplorer.utilities import transformMatrix, applyFdr, getPearson
from hicexplorer._version import __version__


def parse_arguments(args=None):
    """
    parse arguments
    """

    parent_parser = hicexplorer.parserCommon.getParentArgParse()
    parser = argparse.ArgumentParser(
        parents=[parent_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Applies a transformation to a Hi-C matrix.')

    parser.add_argument('--originalMat',
                        help='File name containing the original hic matrix',
                        type=argparse.FileType('r'))

    parser.add_argument('--outFileName', '-o',
                        help='File name to save the resulting matrix ',
                        type=hicexplorer.parserCommon.writableFile,
                        required=True)

    parser.add_argument(
        '--method',
        help='method to transform the matrix values',
        choices=['z-score', 't-score', 'residuals', 'obs/exp', 'p-value',
                 'pearson', 'nbinom-p-value', 'nbinom-expected',
                 'nbinom-est-dist', 'log-norm', 'chi-squared', 'none'],
        required=True)

    parser.add_argument(
        '--applyFdr',
        help='If set the computed values are corrected using fdr. '
        'Should only be used with the methods that produce pvalues.',
        action='store_true')

    parser.add_argument(
        '--perChromosome',
        help='Default is to fit distributions per each distance. Setting this '
        'option will fit distributions per distance per chromosome',
        action='store_true')

    parser.add_argument(
        '--skipDiagonal', '-s',
        help='If set, diagonal counts are not included',
        action='store_true')

    parser.add_argument(
        '--outFormat',
        help='Output format',
        choices=['hdf5', 'dekker'],
        default='hdf5')

    parser.add_argument(
        '--chromosomes',
        help='List of chromosomes to be included in transformation '
        'correction.',
        default=None,
        nargs='+')

    parser.add_argument(
        '--depth',
        help='Depth (in base pairs) up to which the computations will be carried out. A depth of 10.0000 bp '
             'means that any computations involving points that are over 10kbp are not considered.',
        type=int,
        default=None)

    parser.add_argument('--version', action='version',
                        version='%(prog)s {}'.format(__version__))

    return parser


def main():
    """
    collects all arguments and executes
    the appropriate functions
    """
    args = parse_arguments().parse_args()

    hic_ma = HiCMatrix.hiCMatrix(args.matrix)
    if args.originalMat:
        orig_ma = HiCMatrix.hiCMatrix(args.originalMat.name)
    else:
        orig_ma = None

    try:
        hic_ma.maskBins(hic_ma.nan_bins)
    except AttributeError:
        pass

    # remove unwanted Chrs or select a given chromosome
    # in case is given
    hic_ma.filterUnwantedChr()
    if args.originalMat:
        orig_ma.filterUnwantedChr()
    if args.chromosomes:
        hic_ma.keepOnlyTheseChr(args.chromosomes)
        if args.originalMat:
            orig_ma.reorderChromosomes(hic_ma.chrBinBoundaries.keys())

    if args.skipDiagonal:
        hic_ma.diagflat()
        if args.originalMat:
            orig_ma.diagflat()

    max_depth_in_bins = None
    if args.depth:
        binsize = hic_ma.getBinSize()
        if args.depth < binsize:
            exit("Please specify a depth larger than bin size ({})".format(binsize))
        max_depth_in_bins = int(args.depth / binsize)
        import scipy.sparse
        # work only with the upper matrix
        # and remove all pixels that are beyond
        # max_depth_in_bis
        # (this is done by subtracting a second sparse matrix
        # that contains only the upper matrix that wants to be removed.
        hic_ma.matrix = scipy.sparse.triu(hic_ma.matrix, k=0, format='csr') - \
            scipy.sparse.triu(hic_ma.matrix, k=max_depth_in_bins, format='csr')
        hic_ma.matrix.eliminate_zeros()

    if args.method == 'obs/exp':
        hic_ma.convert_to_obs_exp_matrix()
        new_ma = hic_ma.matrix
    elif args.method == 'pearson':
        sys.stderr.write("\nComputing observed / expected\n")
        hic_ma.convert_to_obs_exp_matrix()
        sys.stderr.write("\nComputing pearson\n")
        new_ma = getPearson(hic_ma.matrix)
    elif args.method != 'none':
        # check that the normalized and original matrices
        # have the same size
        if orig_ma:
            assert np.all(hic_ma.matrix.shape == orig_ma.matrix.shape), \
                "original and derived matrices do not have same shape"
        new_ma = transformMatrix(hic_ma, args.method,
                                 per_chr=args.perChromosome,
                                 original_matrix=orig_ma,
                                 depth_in_bins=max_depth_in_bins)
    else:
        new_ma = hic_ma.matrix

    if args.applyFdr:
        new_ma = applyFdr(new_ma)

    hic_ma.setMatrixValues(new_ma)
    hic_ma.restoreMaskedBins()
    if args.outFormat == 'dekker':
        hic_ma.save_dekker(args.outFileName)
    else:
        hic_ma.save(args.outFileName)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import sys
from hicexplorer._version import __version__


def parse_arguments():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""
HiCexplorer addresses the common tasks of Hi-C analysis from processing to visualization.
Each tool should be called by its own name as in the following example:

 $ hicPlotMatrix -m hic_matrix.h5 -o plot.pdf

If you find HiCExplorer useful for your research please cite as:

Fidel Ramirez, Vivek Bhardwaj, Jose Villaveces, Laura Arrigoni, Bjoern A Gruening,Kin Chung Lam,
Bianca Habermann, Asifa Akhtar, Thomas Manke.
“High-resolution TADs reveal DNA sequences underlying genome organization in flies”.
bioRxiv 115063 doi: https://doi.org/10.1101/115063

The following is the list of tools:

   findRestSites            Identifies the genomic locations of restriction sites
   hicBuildMatrix           Creates a Hi-C matrix using the aligned BAM files of the Hi-C sequencing reads
   hicQC                    Plots QC measures from the output of hicBuildMatrix
   hicCorrectMatrix         Uses iterative correction to remove biases from a Hi-C matrix
   hicFindEnrichedContacts  Identifies enriched Hi-C contacts
   hicCorrelate             Computes and visualises the correlation of Hi-C matrices
   hicFindTADs	            Identifies Topologically Associating Domains (TADs)
   hicMergeMatrixBins	    Merges consecutives bins on a Hi-C matrix to reduce resolution
   hicPlotDistVsCounts	    Plot the decay in interaction frequency with distance
   hicPlotMatrix	        Plots a Hi-C matrix as a heatmap
   hicPlotTADs	            Plots TADs as a track that can be combined with other tracks (genes, signal, interactions)
   hicSumMatrices	        Adds Hi-C matrices of the same size
   hicPlotDistVsCounts	    Plots distance vs. Hi-C counts of corrected data
   hicExport	            Export matrix to text formats
   hicInfo                  Shows information about a hicMatrix (no. of bins, bin length, sum, max, min, etc)
   hicCompareMatrices       Computes difference or ratio between two matrices

For more information visit: http://hicexplorer.readthedocs.org
""")

    parser.add_argument('--version', action='version',
                        version='%(prog)s {}'.format(__version__))

    return parser


def process_args(args=None):
    args = parse_arguments().parse_args(args)

    return args


def main(args=None):
    if args is None and len(sys.argv) == 1:
        args = ["--help"]
    process_args(args)
